import logging
import os
import sys
import textwrap
import traceback as tb
import warnings

LOG_FILE = "ragavi.log"
ROOT_LOGGER_LEVEL = "DEBUG"
ROOT_HANDLER_LEVEL = "DEBUG"

# set up the general logger
logger = logging.getLogger("ragavi")


def wrap_warning_text(message, category, filename, lineno, file=None,
                      line=None):
    wrapper = textwrap.TextWrapper(initial_indent=''.rjust(51),
                                   break_long_words=True,
                                   subsequent_indent=''.rjust(49),
                                   width=160)
    message = wrapper.fill(str(message))
    return "%s:%s:\n%s:\n%s" % (filename, lineno,
                                category.__name__.rjust(64), message)


def _handle_uncaught_exceptions(extype, exval, extraceback):
    """Function to Capture all uncaught exceptions into the log file

       Parameters to this function are acquired from sys.excepthook. This
       is because this function overrides :obj:`sys.excepthook`
      `Sys module excepthook <https://docs.python.org/3/library/sys.html#sys.excepthook>`_

    """
    trace = tb.format_exception(extype, exval, extraceback)
    trace = " ".join(trace).split("\n")

    logger.error("Oops ... uncaught exception occurred!")

    for _ in trace:
        if _ != "":
            logger.error(_)


def config_root_logger():
    # capture only a single instance of a matching repeated warning
    warnings.filterwarnings("module")

    # capture warnings from all modules
    logging.captureWarnings(True)

    try:
        cols, rows = os.get_terminal_size(0)
    except:
        # for python2
        cols, rows = (100, 100)

    # only get data from ragavi modules
    logger_filter = logging.Filter("ragavi")

    # setup root logger and defaults
    root_logger = logging.getLogger()
    root_logger.setLevel(ROOT_LOGGER_LEVEL)

    # warnings logger
    w_logger = logging.getLogger("py.warnings")
    w_logger.setLevel("WARNING")

    # console handler
    c_handler = logging.StreamHandler()
    f_handler = logging.FileHandler(LOG_FILE)

    c_handler.setLevel(ROOT_HANDLER_LEVEL)
    f_handler.setLevel(ROOT_HANDLER_LEVEL)

    f_formatter = logging.Formatter(
        "%(asctime)s - %(name)-20s - %(levelname)-10s - %(message)s",
        datefmt="%d.%m.%Y@%H:%M:%S")

    c_handler.setFormatter(f_formatter)
    c_handler.addFilter(logger_filter)

    f_handler.setFormatter(f_formatter)
    f_handler.addFilter(logger_filter)

    root_logger.addFilter(logger_filter)
    root_logger.addHandler(c_handler)
    root_logger.addHandler(f_handler)

    w_logger.addHandler(f_handler)

config_root_logger()
sys.excepthook = _handle_uncaught_exceptions
warnings.formatwarning = wrap_warning_text



import logging

from argparse import ArgumentParser, ArgumentError
from psutil import cpu_count, virtual_memory
from ragavi import __version__

logger = logging.getLogger(__name__)


class ArgumentParserError(Exception):
    pass


class MyParser(ArgumentParser):

    def error(self, message):
        raise ArgumentParserError(message)


def resource_defaults():
    # Value of 1GB
    _GB_ = 2**30

    # setting memory limit in GB
    ml = 1

    # get size of 90% of the RAM available in GB
    mems = virtual_memory()

    logger.info(f"Total RAM size: ~{(mems.total / _GB_):.2f} GB")
    total_mem = int((mems.total * 0.9) / _GB_)

    # set cores to half the amount available
    cores = cpu_count()
    logger.info(f"Total number of Cores: {cores}")
    cores = cores / 2

    if cores > 10:
        cores = 10

    # Because memory is assigned per core
    if (cores * ml) >= total_mem:
        cores = total_mem // ml

    return cores, ml


# for ragavi-vis
def vis_argparser():
    """Create command line arguments for ragavi-vis

    Returns
    -------
    parser : :obj:`ArgumentParser()`
        Argument parser object that contains command line argument's values

    """
    x_choices = ["ant1", "antenna1", "ant2", "antenna2",
                 "amp", "amplitude", "chan", "channel", "freq",  "frequency",
                 "imag", "imaginary", "phase", "real", "scan", "time",
                 "uvdist", "UVdist", "uvdistance", "uvdistl", "uvdist_l",
                 "UVwave", "uvwave"]

    y_choices = ["amp", "amplitude", "imag", "imaginary", "phase", "real"]

    iter_choices = ["ant", "antenna", "ant1", "antenna1", "ant2", "antenna2",
                    "bl", "baseline", "corr", "field", "scan", "spw",
                    #"chan",
                    ]

    cores, ml = resource_defaults()

    parser = MyParser(usage="ragavi-vis [options] <value>",
                      description="A Radio Astronomy Visibilities Inspector")
    parser.add_argument("-v", "--version", action="version",
                        version=f"ragavi {__version__}")

    required = parser.add_argument_group("Required arguments")
    required.add_argument("--ms", dest="mytabs",
                          nargs='+', type=str, metavar='',
                          help="MS to plot. Default is None",
                          default=[])
    required.add_argument('-x', "--xaxis", dest="xaxis", type=str, metavar='',
                          choices=x_choices, help="""X-axis to plot. See https://ragavi.readthedocs.io/en/dev/vis.html#ragavi-vis for the accepted values.""",
                          default=None, required=True)
    required.add_argument("-y", "--yaxis", dest="yaxis", type=str, metavar='',
                          choices=y_choices, help="Y-axis to plot",
                          default=None, required=True)

    pconfig = parser.add_argument_group("Plot settings")

    pconfig.add_argument("-ch", "--canvas-height", dest="c_height", type=int,
                         metavar='',
                         help="""Set height resulting image. Note: This is
                         not the plot height. Default is 720""",
                         default=None)
    pconfig.add_argument("-cw", "--canvas-width", dest="c_width", type=int,
                         metavar='',
                         help="""Set width of the resulting image. Note: This
                        is not the plot width. Default is 1080.""",
                         default=None)
    pconfig.add_argument("--cmap", dest="mycmap", type=str, metavar='',
                         help="""Colour or colour map to use.A list of valid
                         cmap arguments can be found at:
                        https://colorcet.pyviz.org/user_guide/index.html
                        Note that if the argument "colour-axis" is supplied,
                        a categorical colour scheme will be adopted. Default
                        is blue. """,
                         default=None)
    pconfig.add_argument("--cols", dest="n_cols", type=int, metavar='',
                         help="""Number of columns in grid if iteration is
                         active. Default is 9.""",
                         default=9)
    pconfig.add_argument("-ca", "--colour-axis", dest="colour_axis", type=str,
                         metavar='',
                         choices=iter_choices,
                         help="""Select column to colourise by. This will
                         result in a single image. Default is None.""",
                         default=None)
    pconfig.add_argument("--debug", dest="debug",
                         action="store_true",
                         help="""Enable debug messages""")
    pconfig.add_argument("-ia", "--iter-axis", dest="iter_axis", type=str,
                         metavar='', choices=iter_choices,
                         help="""Select column to iterate over. This will
                        result in a grid. Default is None.""",
                         default=None)
    pconfig.add_argument("-lf", "--logfile", dest="logfile", type=str,
                         metavar="",
                         help="""The name of resulting log file (with 
                         preferred extension) If no file extension is 
                         provided, a '.log' extension is appended. The 
                         default log file name is ragavi.log""",
                         default=None)
    pconfig.add_argument("-o", "--htmlname", dest="html_name", type=str,
                         metavar='',
                         help="Output HTML file name (without '.html')",
                         default=None)

    d_config = parser.add_argument_group("Data Selection")
    d_config.add_argument("-a", "--ant", dest="ants", type=str,
                          metavar='',
                          help="""Select baselines where ANTENNA1 corresponds
                          to the supplied antenna(s). "Can be
                        specified as e.g. "4", "5,6,7", "5~7" (inclusive
                        range), "5:8" (exclusive range), 5:(from 5 to last).
                        Default is all.""",
                          default=None)
    d_config.add_argument("--chan", dest="chan", type=str, metavar='',
                          help="""Channels to select. Can be specified using
                        syntax i.e "0:5" (exclusive range) or "20" for
                        channel 20 or "10~20" (inclusive range) (same as
                        10:21) "::10" for every 10th channel or "0,1,3" etc.
                        Default is all.""",
                          default=None)
    d_config.add_argument("-c", "--corr", dest="corr", type=str, metavar='',
                          help="""Correlation index or subset to plot. Can be
                        specified using normal python slicing syntax i.e
                        "0:5" for 0<=corr<5 or "::2" for every 2nd corr or
                        "0" for corr 0  or "0,1,3". Can also be specified
                        using comma separated corr labels e.g 'xx,yy' or
                        specifying 'diag' / 'diagonal' for diagonal
                        correlations and 'off-diag' / 'off-diagonal' for of
                        diagonal correlations. Default is all.""",
                          default=None)
    d_config.add_argument("-dc", "--data-column", dest="data_column",
                          type=str, metavar='',
                          help="""MS column to use for data.
                          Default is DATA.""", default="DATA")
    d_config.add_argument("--ddid", dest="ddid", type=str,
                          metavar='',
                          help="""DATA_DESC_ID(s) /spw to select. Can be
                        specified as e.g. "5", "5,6,7", "5~7" (inclusive
                        range), "5:8" (exclusive range), 5:(from 5 to last).
                        Default is all.""",
                          default=None)
    d_config.add_argument("-f", "--field", dest="fields", type=str,
                          metavar='',
                          help="""Field ID(s) / NAME(s) to plot. Can be
                        specified as "0", "0,2,4", "0~3" (inclusive range),
                        "0:3" (exclusive range), "3:" (from 3 to last) or
                        using a field name or comma separated field names.
                        Default is all""",
                          default=None)
    d_config.add_argument("-if", "--include-flagged", dest="flag",
                          action="store_false",
                          help="Include flagged data in the plot. (Plots both flagged and unflagged data.)",
                          default=True)
    d_config.add_argument("-s", "--scan", dest="scan", type=str, metavar='',
                          help="Scan Number to select. Default is all.",
                          default=None)
    d_config.add_argument("--taql", dest="where", type=str, metavar='',
                          help="TAQL where", default=None)

    d_config.add_argument("--xmin", dest="xmin", type=float, metavar='',
                          help="Minimum x value to plot", default=None)
    d_config.add_argument("--xmax", dest="xmax", type=float, metavar='',
                          help="Maximum x value to plot", default=None)
    d_config.add_argument("--ymin", dest="ymin", type=float, metavar='',
                          help="Minimum y value to plot", default=None)
    d_config.add_argument("--ymax", dest="ymax", type=float, metavar='',
                          help="Maximum y value to plot", default=None)

    avconfig = parser.add_argument_group("Averaging settings")
    avconfig.add_argument("--cbin", dest="cbin", type=int, metavar="",
                          help="""Size of channel bins over which to average
                        .e.g setting this to 50 will average over every 5
                        channels""",
                          default=None)
    avconfig.add_argument("--tbin", dest="tbin", type=float, metavar='',
                          help="""Time in seconds over which to average .e.g
                        setting this to 120.0 will average over every 120.0
                        seconds""",
                          default=None)

    r_config = parser.add_argument_group("Resource configurations")
    r_config.add_argument("-cs", "--chunks", dest="chunks", type=str,
                          metavar='',
                          help="""Chunk sizes to be applied to the dataset.
                          Can be an integer e.g "1000", or a comma separated
                         string e.g "1000,100,2" for multiple dimensions.
                         The available dimensions are (row, chan, corr)
                         respectively. If an integer, the specified chunk
                         size will be applied to all dimensions. If comma
                         separated string, these chunk sizes will be applied
                         to each dimension respectively. Default is 5,000
                         in the row axis.""",
                          default=None)
    r_config.add_argument("-ml", "--mem-limit", dest="mem_limit",
                          type=str, metavar='',
                          default=f"{ml}GB",
                          help="""Memory limit per core e.g '1GB' or '128MB'.
                         Default is 1GB""")
    r_config.add_argument("-nc", "--num-cores", dest="n_cores", type=int,
                          metavar='',
                          help="""Number of CPU cores to be used by Dask.
                        Default is 10 cores. Unless specified, however, this 
                        value may change depending on the amount of RAM on 
                        this machine to ensure that:
                        num-cores * mem-limit < total RAM available""",
                          default=cores)
    return parser


def gains_argparser():
    """Create command line arguments for ragavi-gains

    Returns
    -------
    parser : :obj:`ArgumentParser()`
        Argument parser object that contains command line argument's values

    """
    parser = MyParser(usage="%(prog)s [options] <value>",
                      description="Radio Astronomy Gains Inspector")
    parser.add_argument("-v", "--version", action="version",
                        version=f"ragavi {__version__}")

    required = parser.add_argument_group("Required arguments")
    required.add_argument("-t", "--table", dest="mytabs",
                          nargs='+', type=str, metavar=(' '), required=True,
                          help="""Table(s) to plot. Multiple tables can be
                          specified as a space separated list""",
                          default=[])

    d_config = parser.add_argument_group("Data Selection")
    d_config.add_argument("-a", "--ant", dest="plotants", type=str,
                          metavar='',
                          help="""Plot only a specific antenna, or
                        comma-separated list of antennas. Defaults to all.""",
                          default=None)
    d_config.add_argument("-c", "--corr", dest="corr", type=str, metavar='',
                          help="""Correlation index to plot. Can be a single
                        integer or comma separated integers e.g '0,2'.
                        Defaults to all.""",
                          default=None)
    d_config.add_argument("--ddid", dest="ddid", type=str, metavar='',
                          help="""SPECTRAL_WINDOW_ID or ddid number.
                         Defaults to all""",
                          default=None)
    d_config.add_argument("-f", "--field", dest="fields", type=str, nargs='+',
                          metavar='',
                          help="""Field ID(s) / NAME(s) to plot. Can be
                        specified as "0", "0,2,4", "0~3" (inclusive range),
                        "0:3" (exclusive range), "3:" (from 3 to last) or
                        using a field name or comma separated field names.
                        Defaults to all""",
                          default=None)
    d_config.add_argument("--t0", dest="t0", type=float, metavar='',
                          help="""Minimum time to plot [in seconds].
                        Defaults to full range]""",
                          default=None)
    d_config.add_argument("--t1", dest="t1", type=float, metavar='',
                          help="""Maximum time to plot [in seconds].
                        Defaults to full range""",
                          default=None)
    d_config.add_argument("--taql", dest="where", type=str, metavar='',
                          help="TAQL where clause",
                          default=None)

    pconfig = parser.add_argument_group("Plot settings")
    pconfig.add_argument("--cmap", dest="mycmap", type=str, metavar='',
                         help="""Bokeh or Colorcet colour map to use for 
                         antennas. List of available colour maps can be 
                        found at: https://docs.bokeh.org/en/latest/docs/reference/palettes.html or 
                        https://colorcet.holoviz.org/user_guide/index.html .
                        Defaults to coolwarm""",
                         default="coolwarm")

    pconfig.add_argument("-d", "--doplot", dest="doplot", type=str,
                         metavar='',
                         choices=["ap", "ri"],
                         help="""Plot complex values as amplitude & phase 
                         (ap) or real and imaginary (ri). Defaults to ap.""",
                         default="ap")
    pconfig.add_argument("--debug", dest="debug",
                         action="store_true",
                         help="""Enable debug messages""")
    pconfig.add_argument("-g", "--gaintype", nargs='*', type=str,
                         metavar=' ', dest="gain_types",
                         choices=['B', 'D', 'G', 'K', 'F'],
                         help="""Type of table(s) to be plotted. Can be 
                          specified as a single character e.g. "B" if a 
                          single table has been provided or space 
                          separated list e.g B D G if multiple tables have 
                          been specified. Valid choices are  B D G K & F""",
                         default=[])
    pconfig.add_argument("-kx", "--k-xaxis", dest="kx", type=str, metavar='',
                         choices=["time", "antenna"],
                         help="""Choose the x-xaxis for the K table. Valid 
                        choices are: time or antenna. Defaults to time.""",
                         default="time")
    pconfig.add_argument("-lf", "--logfile", dest="logfile", type=str,
                         metavar="",
                         help="""The name of resulting log file (with 
                         preferred extension) If no file extension is 
                         provided, a '.log' extension is appended. The 
                         default log file name is ragavi.log""",
                         default=None)
    pconfig.add_argument("-o", "--htmlname", dest="html_name", type=str,
                         metavar='',
                         help="""Name of the resulting HTML file. The '.html' 
                        prefix will be appended automatically.""",
                         default=None)
    pconfig.add_argument("-p", "--plotname", dest="image_name", type=str,
                         metavar='', help="""Static image name. The suffix of 
                         this name determines the type of plot. If 
                         foo.png, the output will be PNG, else if foo.svg, 
                         the output will be of the SVG format.""",
                         default=None)

    return parser



import glob
import logging
import numpy as np
import re
import sys
import os

from collections import OrderedDict, namedtuple
from datetime import datetime

import dask.array as da
import daskms as xm
import xarray as xr

from dask import delayed, compute
from bokeh.io import (export_png, export_svgs, output_file, output_notebook,
                      save, show)
from bokeh.layouts import column, grid, gridplot, layout, row
from bokeh.models import (Button, CheckboxGroup,
                          ColumnDataSource, CustomJS, HoverTool,
                          Legend, LinearAxis, Toolbar, PrintfTickFormatter,
                          Slider, Scatter, Toggle, Whisker)

from bokeh.models.widgets import DataTable, TableColumn, Div, PreText
from itertools import product
from pyrap.tables import table

import ragavi.utils as vu
from ragavi.plotting import create_bk_fig, add_axis

# defining some constants
_FLAG_DATA_ = True

_PLOT_WIDTH_ = 900
_PLOT_HEIGHT_ = 710

_GHZ_ = 1e9
_NB_RENDER_ = False
_BATCH_SIZE_ = 16


logger = logging.getLogger(__name__)


#################### Define Dataprocessor ##############################

class DataCoreProcessor:
    """Process gain table data into forms desirable for visualisation. This class is responsible for:
        * x and y data column selection
        * x and y label generation
        * Preparation of x and y data into ColumnDataSource friendly forms
        * Flagging


    Parameters
    ----------
    xds_table_obj : :obj:`xarray.Dataset`
        Table object from which data will be extracted
    ms_name : :obj:`str`
        Name / path to the table being plotted
    gtype : :obj:`str`
        Gain table type
    yaxis: :obj:`str`
        Desired y-axis being plotted
    corr : :obj:`int`
        Correlation index being plotted
    chan : :obj:`slice`
        A slicer to select channels
    ddid : :obj:`int`
        SPW id fo data being plotted
    flag : :obj:`bool`
        To flag or not. Default is True.
    fid : :obj:`int`
        Field id being plotted
    kx : :obj:`str`
        Applicable to delay (K) tables. Determines the x-axis of the plot. Default is time.
    """

    def __init__(self, xds_table_obj, ms_name, gtype, yaxis, chan=None,
                 corr=0, ddid=None, fid=None, flag=None, kx=None):

        self.xds_table_obj = xds_table_obj
        self.ms_name = ms_name
        self.gtype = gtype
        self.fid = fid
        self.chan = chan
        self.corr = corr
        self.ddid = ddid
        self.flag = flag
        self.kx = kx
        self.xaxis = self.set_xaxis()
        self.yaxis = yaxis

    def __repr__(self):
        return "DataCoreProcessor(%r, %r, %r, %r)" % (
            self.xds_table_obj, self.ms_name, self.gtype, self.yaxis)

    def set_xaxis(self):
        if self.gtype in ['B', 'D']:
            xaxis = "channel"
        elif self.gtype in ['G', 'F']:
            xaxis = "time"
        elif self.gtype == "K":
            if self.kx == "time":
                xaxis = "time"
            else:
                xaxis = "antenna"
        return xaxis

    def process_data(self, ydata, yaxis=None):
        """Abstraction for processing y-data passes it to the processing function.

        Parameters
        ----------
        ydata: :obj:`xarray.DataArray`
               y-data to process

        Returns
        -------
        y: :obj:`xarray.DataArray`
           Processed :attr:`ydata`
        """
        if yaxis is None:
            yaxis = self.yaxis

        if yaxis == "amplitude":
            y = vu.calc_amplitude(ydata)
        elif yaxis == "imaginary":
            y = vu.calc_imaginary(ydata)
        elif yaxis == "phase":
            y = vu.calc_phase(ydata, unwrap=False)
        elif yaxis == "real":
            y = vu.calc_real(ydata)
        elif yaxis == "delay" or yaxis == "error":
            y = ydata

        return y

    def flatten_bandpass_data(self, y, x=None):
        """ Massage B table with multiple unique times into a shape
        that works with BokehJS. Returns on x if x input is present
        otherwise returns y.

        Parameters
        ----------
        x : :obj:`np.array`
            1-D Array containing data for the x-axis
        y : :obj:`np.array`
            2-D Array containing y-axis data.

        Returns
        -------
        x : :obj:`np.array`
            1-D Array containing repeated data for the x-axis
        y : :obj:`np.array`
            1-D Array containing y-axis flattened data.


        """
        logger.debug(f"Flattening bandpass: In y shape {str(y.shape)}")
        if y.ndim > 1:
            # no of unique times available in the bandpass
            # last index because y is transposed to (chan, time)
            nrows = y.shape[1]

        if x is not None:
            logger.debug(f"Flattening bandpass: In x shape {str(x.shape)}")
            x = np.tile(x, (nrows))
            logger.debug(f"Out x shape: {str(x.shape)}")
            return x
        else:
            # TODO check if this flattening works proper
            y = y.T.flatten()
            logger.debug(f"Out y shape: {str(y.shape)}")
            return y

    def get_errors(self):
        """Get error data from PARAMERR column.
        Returns
        -------
        errors: :obj:`xarray.DataArray`
            Data from the PARAMERR column
        """
        errors = self.xds_table_obj.PARAMERR
        return errors

    def get_xaxis_data(self):
        """Get x-axis data. This function also returns the relevant x-axis label

        Returns
        -------
        xdata : :obj:`xarray.DataArray`
            X-axis data depending  x-axis selected.
        x_label : :obj:`str`
            Label to appear on the x-axis of the plots.
        """
        if self.xaxis in ["antenna", "antenna1"]:
            xdata = self.xds_table_obj.ANTENNA1
            x_label = "Antenna"
        elif self.xaxis == "channel":
            xdata = vu.get_frequencies(self.ms_name, spwid=self.ddid,
                                       chan=self.chan)
            x_label = "Channel"
        elif self.xaxis == "time":
            xdata = self.xds_table_obj.TIME
            x_label = "Time"
        else:
            logger.error("Invalid x-axis name")
            return

        return xdata, x_label

    def prep_xaxis_data(self, xdata, freq=None):
        """Prepare the x-axis data for plotting. Selections also performed here.

        Parameters
        ----------
        xdata: :obj:`xarray.DataArray`
            x-axis data depending  x-axis selected

        freq: :obj:`xarray.DataArray` or :obj:`float`
            Frequency(ies) from which corresponding wavelength will be obtained.
            Note
            ----
                Only required when xaxis specified is "uvwave"

        Returns
        -------
        prepdx: :obj:`xarray.DataArray`
            Prepared :attr:`xdata`
        """
        if self.xaxis in ["channel", "frequency"]:
            prepdx = xdata.chan
        elif self.xaxis == "time":
            prepdx = vu.time_convert(xdata)
        elif self.xaxis in ["antenna1", "antenna"]:
            prepdx = np.full((self.xds_table_obj.row.size), (xdata))

        logger.debug(f"xaxis: {self.xaxis} data ready")
        return prepdx

    def get_yaxis_data(self):
        """Extract the required column for the y-axis data.

        Returns
        -------
        ydata, y_label : :obj:`xarray.DataArray`, :obj:`str`
                         :attr:`ydata` containing y-axis data depending  y-axis selected and `y_label` which is the label to appear on the y-axis of the plots.
        """

        # default data column
        datacol = "CPARAM"

        if self.yaxis == "amplitude":
            y_label = "Amplitude"
        elif self.yaxis == "imaginary":
            y_label = "Imaginary"
        elif self.yaxis == "phase":
            y_label = "Phase [deg]"
        elif self.yaxis == "real":
            y_label = "Real"
        elif self.yaxis == "delay":
            datacol = "FPARAM"
            y_label = "Delay[ns]"
        # attempt to get the specified column from the table
        try:
            ydata = self.xds_table_obj[datacol]
        except KeyError:
            logger.exception("Column '{}'' not Found".format(datacol))
            return sys.exit(-1)
        return ydata, y_label

    def prep_yaxis_data(self, ydata):
        """Process data for the y-axis.

        This function is responsible for :

            * Correlation selection
            * Flagging
            * Conversion form complex to the required form
            * Data transposition if `yaxis` is frequency or channel

        Data selection and flagging are done by this function itself, however ap and ri conversion are handled by :meth:`ragavi.ragavi.DataCoreProcessor.compute_ydata`

        Parameters
        ----------
        ydata : :obj:`xarray.DataArray`
            y-axis data to be processed
        Returns
        -------
        y : :obj:`xarray.DataArray`
            Processed :attr:`ydata`
        """
        ydata = ydata.sel(corr=self.corr)

        if self.flag:
            logger.debug("Flagging is active")
            flags = vu.get_flags(self.xds_table_obj)
            flags = flags.sel(corr=self.corr)

            # if flagging is activated select data only where flag mask is 0
            processed = self.process_data(ydata)
            y = processed.where(flags == False)
        else:
            y = self.process_data(ydata)

        if self.xaxis == "channel":
            y = y.T

        logger.debug(f"y-axis: {self.yaxis }, data ready")
        return y

    def blackbox(self):
        """ Get raw input data and churn out processed data. Takes in all inputs from the instance initialising object

        This function incorporates all function in the class to get the desired result. It performs:

            - xaxis data and error data acquisition
            - xaxis data and error preparation and processing
            - yaxis data and error data acquisition
            - yaxis data and error preparation and processing

        Returns
        -------
        d : :obj:`collections.namedtuple`
            A named tuple containing all processed x-axis data, errors and label, as well as both pairs of y-axis data, their error margins and labels. Items from this tuple can be gotten by using the dot notation.

        """
        logger.debug(vu.ctext("Blackbox in"))
        logger.debug("DCP: blackbox: Getting x-axis data")

        Data = namedtuple("Data", "x x_label y y_label y_err")
        x = self.x_only()
        xdata = x.x
        xlabel = x.x_label

        logger.debug("DCP: blackbox: Getting y-axis data")
        y = self.y_only()
        ydata = y.y
        ylabel = y.y_label
        hi, lo = y.y_err

        if self.xaxis == "channel":
            xdata = self.flatten_bandpass_data(ydata, x=xdata)
            ydata = self.flatten_bandpass_data(ydata)
            hi = self.flatten_bandpass_data(hi)
            lo = self.flatten_bandpass_data(lo)

        xdata, ydata, hi, lo = compute(xdata, ydata, hi, lo)

        d = Data(x=xdata, x_label=xlabel, y=ydata,
                 y_label=ylabel, y_err=(hi, lo))

        logger.debug(vu.ctext("Blackbox out"))
        return d

    def act(self):
        """Activate the :meth:`ragavi.ragavi.DataCoreProcessor.blackbox`
        """
        return self.blackbox()

    def x_only(self):
        """Return only x-axis data and label

        Returns
        -------
        d : :obj:`collections.namedtuple`
            Named tuple containing x-axis data and x-axis label. Items in the tuple can be accessed by using the dot notation.
        """
        Data = namedtuple("Data", "x x_label")
        xdata, xlabel = self.get_xaxis_data()
        xdata = self.prep_xaxis_data(xdata)
        if not isinstance(xdata, np.ndarray):
            xdata = xdata.data
        d = Data(x=xdata, x_label=xlabel)
        return d

    def y_only(self, add_error=True):
        """Return only y-axis data and label

        Returns
        -------
        d : :obj:`collections.namedtuple`
            Named tuple containing x-axis data and x-axis label. Items in the tuple can be accessed by using the dot notation.
        """

        Data = namedtuple("Data", "y y_label y_err")
        ydata, y_label = self.get_yaxis_data()
        p_ydata = self.prep_yaxis_data(ydata)
        if add_error:
            err_data = self.get_errors()
            hi = self.prep_yaxis_data(ydata + err_data)
            lo = self.prep_yaxis_data(ydata - err_data)
        else:
            hi = None
            lo = None

        d = Data(y=p_ydata.data, y_label=y_label, y_err=(hi.data, lo.data))
        return d


################### Some table related functions #######################

def get_table(tab_name, antenna=None, fid=None, spwid=None, where=[],
              group_cols=None):
    """ Get xarray Dataset objects containing gain table columns of the selected data

    Parameters
    ----------
    antenna: :obj:`str`, optional
        A string containing antennas whose data will be selected
    fid: :obj:`int`, optional
        FIELD_ID whose data will be selected
    spwid: :obj:`int`, optional
        DATA_DESC_ID or spectral window whose data will be selected
    tab_name: :obj:`str`
        name of your table or path including its name
    where: :obj:`list`, optional
        TAQL where clause to be used within the table.

    Returns
    -------
    tab_objs: :obj:`list`
        A list containing :obj:`xarray.Dataset` objects where each item on the list is determined by how the data is grouped

    """

    logger.debug(f"Reading calibration table: {tab_name}")

    # defining part of the gain table schema
    tab_schema = {"CPARAM": {"dims": ("chan", "corr")},
                  "FLAG": {"dims": ("chan", "corr")},
                  "FPARAM": {"dims": ("chan", "corr")},
                  "PARAMERR": {"dims": ("chan", "corr")},
                  "SNR": {"dims": ("chan", "corr")},
                  }

    # where is now a list
    if antenna != None:
        where.append("ANTENNA1 IN {}".format(antenna))
    if fid != None:
        where.append("FIELD_ID IN {}".format(fid))
    if spwid != None:
        if spwid.isnumeric():
            spwid = int(spwid)
        else:
            spwid = vu.resolve_ranges(spwid)
        where.append("SPECTRAL_WINDOW_ID IN {}".format(spwid))
    if len(where) > 0:
        where = " && ".join(where)
    else:
        where = ""

    if group_cols is None:
        group_cols = ["SPECTRAL_WINDOW_ID", "FIELD_ID", "ANTENNA1"]

    logger.debug(f"Grouping data by: {' '.join(group_cols) or None}")
    logger.debug(f"TAQL selection: {where}")

    try:
        tab_objs, t_keywords = xm.xds_from_table(tab_name, taql_where=where,
                                                 table_schema=tab_schema,
                                                 group_cols=group_cols,
                                                 table_keywords=True)

        # get table type from VisCal
        g_type = t_keywords["VisCal"].split()[0]

        logger.info(f"Table type: {g_type} Jones")

        logger.debug(f"{len(tab_objs)} groups created from table")
        return tab_objs, g_type

    except Exception as ex:
        logger.error(ex, exc_info=True)
        sys.exit(-1)


def get_time_range(tab_name, unix_time=True):
    """Get the initial and final TIME column before selections

    Returns
    -------
    init_time, f_time : :obj:`tuple`
        Containing initial time and final time available in the ms
    """

    # unix_time = MJD - munix
    munix = 3506716800.0

    ms = table(tab_name, ack=False)
    i_time = ms.getcell("TIME", 0)
    f_time = ms.getcell("TIME", (ms.nrows() - 1))

    if unix_time:
        i_time = i_time - munix
        f_time = f_time - munix

    ms.close()

    logger.debug(f"Time range Initial: {str(i_time)}, final: {str(f_time)}")

    return i_time, f_time


def get_tooltip_data(xds_table_obj, gtype, freqs):
    """Get the data to be displayed on the tool-tip of the plots

    Parameters
    ----------
    gtype: :obj:`str`
        Type of gain table being plotted
    xds_table_obj : `xarray.Dataset`
        xarray-ms table object
    freqs : :obj:`np.array`
        An array containing frequencies for the current SPW

    Returns
    -------
    spw_id : :obj:`numpy.ndarray`
        Spectral window ids
    scan_no : :obj:`numpy.ndarray`
        scan ids

    """
    logger.debug("Getting tooltip data")

    scan_no = xds_table_obj.SCAN_NUMBER.values
    scan_no = scan_no.astype(np.uint16)
    spw_id = np.full(scan_no.shape, xds_table_obj.SPECTRAL_WINDOW_ID,
                     dtype=np.uint8)

    # get the number of channels
    nchan = freqs.size

    if gtype in ['B', 'D']:
        # extend array to accommodate for multiple time slots
        spw_id = np.array(spw_id).repeat(nchan, axis=0)
        # tile because of different scan numbers available
        scan_no = np.tile(scan_no, nchan)

    logging.debug("Tooltip data ready")
    return spw_id, scan_no

#################### JS callbacks ######################################
# Visible data selection callbacks


def ant_select_callback():
    """JS callback for the selection and de-selection of antennas

    Returns
    -------
    code : :obj:`str`
    """

    code = """
            /*ax: list containing glyph renderers
              bsel: batch selection group button
              nbatches: Number of batches available
              fsel: field_selector widget,
              csel: corr_select widget,
              ssel: spw_select widget,
              nfields: Number of available fields,
              ncorrs: Number of avaiblable corrs,
              nspws: Number of spectral windows,
            */
            let nplots = ax.length;
            if (cb_obj.active.includes(0))
                {
                    for(let i=0; i<nplots; i++){
                        ax[i].visible = true;
                    }
                    //activate all the checkboxes whose antennas are active
                    bsel.active = [...Array(nbatches).keys()]
                    csel.active = [...Array(ncorrs).keys()]
                    fsel.active = [...Array(nfields).keys()]
                    ssel.active = [...Array(nspws).keys()]

                }
            else{
                    for(let i=0; i<nplots; i++){
                       ax[i].visible = false;
                    }

                    bsel.active = []
                    csel.active = []
                    fsel.active = []
                    ssel.active = []
                }
            """

    return code


def batch_select_callback():
    """JS callback for batch selection Checkboxes

    Returns
    -------
    code : :obj:`str`
    """
    code = """
        /* antsel: Select or deselect all antennas button
           bsize: total number of items in a batch
           csel: the corr selector array
           fsel: the field selector button
           ssel: spw selector button
           nants: Number of antennas
           ncorrs: number of available correlations
           nfields: number of available fields
           nbatches: total number of available batches
           ax: List containing glyphs for all antennas, fields and correlations
           count: keeping a cumulative sum of the traverse number
        */

        let count = 0;
        let new_bsize;
        for (let sp=0; sp<nspws; sp++){
            for (let f=0; f<nfields; f++){
                for (let c=0; c<ncorrs; c++){
                    //re-initialise bsize
                    new_bsize = bsize;
                    for(let n=0; n<nbatches; n++){
                        /* Reduce batch size to the size of the last batch
                        and number of antennas is not the same as bsize
                        */
                        if (n == nbatches-1 && nants!=bsize){
                            new_bsize = nants % bsize;
                        }
                        for(let b=0; b<new_bsize; b++){
                            if (cb_obj.active.includes(n)){
                                ax[count].visible = true;
                                }
                            else{
                                ax[count].visible = false;
                            }
                            count = count + 1;
                        }
                    }
                }
            }
        }

        if (cb_obj.active.length == nbatches){
            antsel.active = [0];
        }
        else if (cb_obj.active.length==0){
            antsel.active = [];
        }
       """
    return code


def corr_select_callback():
    """Correlation selection callback"""
    code = """
         /*bsize: total number of items in a batch
           bsel: the batch selector group buttons
           fsel: field selector group buttons
           ssel: spw selector group buttons
           nants: Total number of antennas
           nspws: Number of spectral windows
           ncorrs: number of available correlations
           nfields: number of available fields
           nbatches: total number of available batches
           ax: List containing glyphs for a single plot for all antennas, fields and correlations
           count: keeping a cumulative sum of the traverse number
        */

        let count = 0;
        let new_bsize;
        for (let sp=0; sp<nspws; sp++){
            for (let f=0; f<nfields; f++){
                for (let c=0; c<ncorrs; c++){
                    //re-initialise new batch size
                    new_bsize = bsize;
                    for(let n=0; n<nbatches; n++){
                        // Reduce new_bsize to the size of the final batch
                        if (n == nbatches-1 && nants!=bsize){
                            new_bsize = nants % bsize;
                        }
                        for(let b=0; b<new_bsize; b++){
                            if (cb_obj.active.includes(c) && fsel.active.includes(f) &&
                                ssel.active.includes(sp)){
                                ax[count].visible = true;
                                }
                            else{
                                ax[count].visible = false;
                            }
                            count = count + 1;
                        }
                    }
                }
            }
        }

       """
    return code


def field_selector_callback():
    """Return JS callback for field selection checkboxes

    Returns
    -------
    code : :obj:`str`
    """
    code = """
        /*bsize: total number of items in a batch
         bsel: the batch selector group buttons
         csel: corr selector group buttons
         nants: Number of antennas
         ncorrs: number of available correlations
         nfields: number of available fields
         nbatches: total number of available batches
         ax: List containing glyphs for all antennas, fields and  correlations
         count: keeping a cumulative sum of the traverse number
        */
        let count = 0;
        let new_bsize;
        for (let sp=0; sp<nspws; sp++){
            for (let f=0; f<nfields; f++){
                for (let c=0; c<ncorrs; c++){
                    //re-initialise new batch size
                    new_bsize = bsize;
                    for(let n=0; n<nbatches; n++){
                        // Reduce new batch size to the size of the last batch
                        if (n == nbatches-1 && nants!=bsize){
                            new_bsize = nants % bsize;
                        }
                        for(let b=0; b<new_bsize; b++){
                            if (cb_obj.active.includes(f) && csel.active.includes(c) &&
                                ssel.active.includes(sp)){
                                ax[count].visible = true;
                                }
                            else{
                                ax[count].visible = false;
                            }
                            count = count + 1;
                        }
                    }
                }
            }
        }
       """
    return code


def spw_select_callback():
    """SPW selection callaback"""
    code = """
         /*bsize: total number of items in a batch
           bsel: the batch selector group buttons
           ex_ax: for the extra frequency ax(es). Is 0 if none.
           fsel: field selector group buttons
           csel: corr selector group buttons
           nspws: Number of spectral windows
           ncorrs: number of available correlations
           nfields: number of available fields
           nbatches: total number of available batches
           ax: List containing glyphs for all antennas, fields and correlations
           count: keeping a cumulative sum of the traverse number
        */

        let count = 0;
        let new_bsize;
        for (let sp=0; sp<nspws; sp++){
            for (let f=0; f<nfields; f++){
                for (let c=0; c<ncorrs; c++){
                    //re-initialise new batch size
                    new_bsize = bsize;

                    for(let n=0; n<nbatches; n++){
                        // Reduce bsize to the size of the last batch
                        if (n == nbatches-1 && nants!=bsize){
                            new_bsize = nants % bsize;
                        }
                        for(let b=0; b<new_bsize; b++){
                            if (cb_obj.active.includes(sp) && fsel.active.includes(f) &&
                                csel.active.includes(c)){
                                ax[count].visible = true;
                                }
                            else{
                                ax[count].visible = false
                            }
                            count = count + 1;
                        }
                    }
                }
            }

            //Make the extra y-axes visible only if corresponding spw selected
            if (cb_obj.active.includes(sp)){
                ex_ax[sp].visible=true;
            }
            else{
                ex_ax[sp].visible=false;

            }

        }

       """
    return code


# Show additional data callbacks

def flag_callback():
    """JS callback for the flagging button

    Returns
    -------
    code : :obj:`str`
    """
    code = """
            //sources: list of all different sources for the different antennas and available sources
            //n_sources: number of sources
            //flagging: status of flag_data

            let n_sources =  uf_sources.length;

            if (cb_obj.active.includes(0)){
                for (let i=0; i<n_sources; i++){
                    uf_sources[i].data.y1 = f_sources[i].data.iy1;
                    uf_sources[i].data.y2 = f_sources[i].data.iy2;
                    uf_sources[i].change.emit();
                }
            }
            else{
                for (let i=0; i<n_sources; i++){
                    uf_sources[i].data.y1 = f_sources[i].data.y1;
                    uf_sources[i].data.y2 = f_sources[i].data.y2;
                    uf_sources[i].change.emit();
                }


            }
           """
    return code


def toggle_err_callback():
    """JS callback for Error toggle Toggle button

    Returns
    -------
    code : :obj:`str`
    """
    code = """
            let n_glyphs = ax.length;
             //if toggle button active
            if (cb_obj.active.includes(0))
                {
                    for(let i=0; i<n_glyphs; i++){
                        //only switch on if corresponding plot is on
                        if(ax[i].visible){
                            ax_err[i].visible = true;
                        }

                    }
                }
            else{
                    for(let i=0; i<n_glyphs; i++){
                        ax_err[i].visible = false;
                    }
                }
            """
    return code


def legend_toggle_callback():
    """JS callback for legend toggle Dropdown menu

    Returns
    -------
    code : :obj:`str`
    """
    code = """
                //Show only the legend for the first item
                let n;
                if (cb_obj.active.includes(0)){
                    for (n=0; n<n_legs; n++){
                        legs[n].visible = true;
                        }
                }

                else{
                    for (n=0; n<n_legs; n++){
                        legs[n].visible = false;
                        }
                }

           """
    return code


# Plot layout callbacks

def size_slider_callback():
    """JS callback to select size of glyphs

    Returns
    -------
    code : :obj:`str`
    """

    code = """
            let n_rs = ax.length;
            for (let i=0; i<n_rs; i++){
                ax[i].size = cb_obj.value;
            }
           """
    return code


def alpha_slider_callback():
    """JS callback to alter alpha of glyphs

    Returns
    -------
    code : :obj:`str`
    """

    code = """
            let n_rs = ax.length;
            for (let i=0; i<n_rs; i++){
                ax[i].fill_alpha = cb_obj.value;
            }
           """
    return code


def axis_fs_callback():
    """JS callback to alter axis label font sizes

    Returns
    -------
    code : :obj:`str`
    """
    code = """
            let n_axes = ax.length;
            for (let i=0; i<n_axes; i++){
                ax[i].axis_label_text_font_size = `${cb_obj.value}pt`;
            }

           """
    return code


def title_fs_callback():
    """JS callback for title font size slider

    Returns
    -------
    code : :obj:`str`
    """
    code = """
            ax.text_font_size = `${cb_obj.value}pt`;
           """
    return code


# Download data selection

def save_selected_callback():
    code = """
        /*uf_src: Unflagged data source
          f_src:  Flagged data source scanid antname
        */
        let out = `x, y1, y2, ant, corr, field, scan, spw\n`;

        //for all the data sources available
        for (let i=0; i<uf_src.length; i++){
            let sel_idx = uf_src[i].selected.indices;
            let data = uf_src[i].data;

            for (let j=0; j<sel_idx.length; j++){
                out +=  `${data['x'][sel_idx[j]]}, ` +
                        `${data['y1'][sel_idx[j]]}, ` +
                        `${data['y2'][sel_idx[j]]}, ` +
                        `${data['antname'][sel_idx[j]]}, ` +
                        `${data['corr'][sel_idx[j]]}, ` +
                        `${data['field'][sel_idx[j]]}, ` +
                        `${data['scanid'][sel_idx[j]]}, ` +
                        `${data['spw'][sel_idx[j]]}\n`;
            }

        }
        let answer = confirm("Download selected data?");
        if (answer){
            let file = new Blob([out], {type: 'text/csv'});
            let element = window.document.createElement('a');
            element.href = window.URL.createObjectURL(file);
            element.download = "data_selection.csv";
            document.body.appendChild(element);
            element.click();
            document.body.removeChild(element);
        }

    """

    return code


####################### Plot Related functions ########################

def condense_legend_items(inlist):
    """Combine renderers of legend items with the same legend labels. Must be
    done in case there are groups of renderers which have the same label due
    to iterations, to avoid a case where there are two or more groups of
    renderers containing the same label name.

    Parameters
    ----------
    inlist : :obj:`list`
             # bokeh.models.annotations.LegendItem>`_
             List containing legend items of the form (label, renders) as
             described in: `Bokeh legend items <https://bokeh.pydata.org/en/latest/docs/reference/models/annotations.html
    Returns
    -------
    outlist : :obj:`list`
              A reduction of :attr:`inlist` containing renderers sharing the
              same labels grouped together.
    """
    od = OrderedDict()
    # put all renderers with the same legend labels together
    for key, value in inlist:
        if key in od:
            od[key].extend(value)
        else:
            od[key] = value

    # reformulate odict to list
    outlist = [(key, value) for key, value in od.items()]

    logger.debug("Grouped renderers with the same legends")

    return outlist


def create_legend_batches(num_leg_objs, li_ax1, batch_size=16):
    """Automates creation of antenna **batches of 16** each unless otherwise.

    This function takes in a long list containing all the generated legend items from the main function's iteration and divides this list into batches, each of size :attr:`batch_size`. The outputs provides the inputs to
    :meth:`ragavi.ragavi.create_legend_objs`.


    Parameters
    ----------
    batch_size : :obj:`int`, optional
        Number of antennas in a legend object. Default is 16
    li_ax1 : :obj:`list`
        # bokeh.models.annotations.LegendItem>`_ for antennas for 1st figure
        # items are in the form (antenna_legend, [renderer])
        List containing all `legend items <https://bokeh.pydata.org/en/latest/docs/reference/models/annotations.html
    num_leg_objs : :obj:`int`
        Number of legend objects to be created

    Returns
    -------
    bax1 : :obj:`list`
           Tuple containing List of lists which each have :attr:`batch_size` number of legend items for each batch.
           bax1 are batches for figure1 antenna legends, and ax2 batches for figure2 antenna legends

           e.g bax1 = [[batch0], [batch1], ...,  [batch_numOfBatches]]
    """

    bax1 = []

    # condense the returned list if two fields / corrs/ spws were plotted
    li_ax1 = condense_legend_items(li_ax1)

    j = 0
    for i in range(num_leg_objs):
        # in case the number is not a multiple of 16 or is <= num_leg_objs
        # or on the last iteration

        if i == num_leg_objs:
            bax1.extend([li_ax1[j:]])
        else:
            bax1.extend([li_ax1[j:j + batch_size]])

        j += batch_size

    logger.debug(f"Ant batches created. Batch size: {str(batch_size)}, Batches: {str(num_leg_objs)}")

    return bax1


def create_legend_objs(num_leg_objs, bax1):
    """Creates legend objects using items from batches list
    Legend objects allow legends be positioning outside the main plot

    Parameters
    ----------
    num_leg_objs: :obj:`int`
        Number of legend objects to be created
    bax1 : :obj:`list`
        Batches for antenna legends for plot

    Returns
    -------
    lo_ax1 : :obj:`dict`
        Dictionaries with legend objects for a plot


    """

    lo_ax1 = {}

    l_opts = dict(click_policy="hide", glyph_height=20,
                  glyph_width=20, label_text_font_size="8pt",
                  label_text_font="monospace", location="top_left",
                  margin=1, orientation="horizontal",
                  level="annotation", spacing=1,
                  padding=2, visible=False)
    for i in range(num_leg_objs):
        leg1 = Legend(items=bax1[i], name=f"leg{i}", **l_opts)
        lo_ax1["leg_%s" % str(i)] = leg1

    logger.debug("Legend objects created")

    return lo_ax1


def errorbar(x, y, yerr=None, color="red"):
    """Add errorbars to Figure object based on :attr:`x`, :attr:`y` and attr:`yerr`

    Parameters
    ----------
    color : :obj:`str`
        Color for the error bars
    x : :obj:`numpy.ndarray`
        x_axis data
    y : :obj:`numpy.ndarray`
        y_axis data
    yerr : :obj:`numpy.ndarray`, :obj:`numpy.ndarray`
        Tuple with high and low limits for :attr:`y`

    Returns
    -------
    ebars : :obj:`bokeh.models.Whisker`
        Return the object containing error bars
    """
    # Setting default return value
    ebars = None

    if yerr is not None:

        src = ColumnDataSource(data=dict(upper=yerr[0],
                                         lower=yerr[1], base=x))

        ebars = Whisker(source=src, base="base", upper="upper",
                        lower="lower", line_color=color, name="p_whisker",
                        visible=False)
        ebars.upper_head.line_color = color
        ebars.lower_head.line_color = color

        logger.debug("Errorbars created")
    return ebars


def gen_flag_data_markers(y, fid=None, markers=None, fmarker="circle_x"):
    """Generate different markers for where data has been flagged.

    Parameters
    ----------
    fid : :obj:`int`
        Field id number to identify the marker to be used
    fmarker : :obj:`str`
        Marker to be used for flagged data
    markers : :obj:`list`
        A list of all available bokeh markers
    y : :obj:`numpy.ndarray`
        The flagged data containing NaNs

    Returns
    -------
    masked_markers_arr : :obj:`numpy.ndarray`
        Returns an n-d array of shape :code:`y.shape` containing markers for valid data and :attr:`fmarker` where the data was NaN.
    """

    # fill an array with the unflagged marker value
    markers_arr = np.full(y.shape, fill_value=markers[fid], dtype="<U17")

    # mask only where there are nan values
    masked_markers_arr = np.ma.masked_where(np.isnan(y), markers_arr)
    # fill with the different marker
    masked_markers_arr.fill_value = fmarker

    # return filled matrix
    masked_markers_arr = masked_markers_arr.filled()

    logger.debug("Flagged data markers generated")

    return masked_markers_arr


def link_plots(all_figures=None, all_fsources=None, all_ebars=None):
    """ Link all plots generated by this script. Done to unify interactivity
    within the plots such as panning, zoomin etc.

    Parameters
    ----------
    all_figures : :obj:`list`
        Containing all the figures generated in the plot
    all_fsources: :obj:`list`
        All data sources containing flagged data for all figures
    all_ebars : :obj:`list`
        All error bars for all figures

    Returns
    -------
    all_fsrc, all_ufsrc : :obj:`tuple`
        Unified flagged and un-flagged data sources
    """
    n_figs = len(all_figures)
    fig1 = all_figures[0]
    fig1_ebars = all_ebars[0]

    # number of axes in the figure
    n_axis = len(fig1.axis)

    # number of renderers
    n_rs = len(fig1.renderers)
    all_ufsrc = []
    all_fsrc = []

    logger.debug("Linking generated plots")

    for f in range(1, n_figs):
        # link the x- ranges
        all_figures[f].x_range = fig1.x_range
        # link the titles font sizes
        fig1.select(name="p_title")[0].js_link(
            "text_font_size", all_figures[f].select(name="p_title")[0],
            "text_font_size")

        # link x and y-axes font sizes
        for _i in range(n_axis):
            fig1.axis[_i].js_link(
                "axis_label_text_font_size", all_figures[f].axis[_i],
                "axis_label_text_font_size")
            fig1.axis[_i].js_link(
                "visible", all_figures[f].axis[_i],
                "visible")

        for _i in range(n_rs):
            # add y data from fig 1
            fig1.renderers[_i].data_source.add(
                all_figures[f].renderers[_i].data_source.data[f"y{f+1}"],
                name=f"y{f+1}")

            # place the data in a single CDS to be shared
            shared_cds = ColumnDataSource(
                data=dict(fig1.renderers[_i].data_source.data))

            # ensure that the cds and its view are the same otherwise error
            all_figures[f].renderers[_i].data_source = shared_cds
            all_figures[f].renderers[_i].view.source = shared_cds
            fig1.renderers[_i].data_source = shared_cds
            fig1.renderers[_i].view.source = shared_cds

            all_fsources[0][_i].add(all_fsources[f][_i].data[f"y{f+1}"],
                                    name=f"y{f+1}")
            all_fsources[0][_i].add(all_fsources[f][_i].data[f"iy{f+1}"],
                                    name=f"iy{f+1}")

            # link visible values for the renderers
            fig1.renderers[_i].js_link("visible",
                                       all_figures[f].renderers[_i],
                                       "visible")
            fig1.renderers[_i].glyph.js_link("size",
                                             all_figures[f].renderers[_i].glyph, "size")
            fig1.renderers[_i].glyph.js_link("fill_alpha",
                                             all_figures[f].renderers[_i].glyph, "fill_alpha")
            if fig1_ebars[_i]:
                fig1_ebars[_i].js_link("visible", all_ebars[f][_i], "visible")
            all_ufsrc.append(shared_cds)
            all_fsrc.append(all_fsources[0][_i])

    return all_fsrc, all_ufsrc


def make_plots(source, fid=0, color="red", yerr=None, yidx=None):
    """Generate a pair of plots

    Parameters
    ----------
    fig : :obj:`bokeh.plotting.figure`
        First figure
    color : :obj:`str`
        Glyph color[s]
    fid : :obj:`int`
        field id number to set the line width
    source : :obj:`bokeh.models.ColumnDataSource`
            Data source for the renderer
    yerr : :obj:`numpy.ndarray`, optional
        Y-axis error margins
    yidx : :obj:`int`
        Current enumerated y-axis number. Used for keeping track of the
        figures.


    Returns
    -------
    p_glyph, ebars: (:obj:`bokeh.models.Glyph`, :obj:`bokeh.models.Whisker`)
    Tuple of containing glyphs and error bars to be used (from :func:`ragavi.ragavi.errorbar`).

    """
    markers = ["circle", "diamond", "square", "triangle",
               "hex"]
    fmarker = "inverted_triangle"
    glyph_opts = {"size": 4,
                  "fill_alpha": 1,
                  "fill_color": color,
                  "line_color": "black",
                  # "nonselection_fill_color": "#7D7D7D",
                  # "nonselection_fill_alpha": 0.3,
                  "name": f"p_glyph_{yidx}"}

    # if there is any flagged data enforce fmarker where flag is active
    fmarkers = None
    if np.any(np.isnan(source.data[f"y{yidx}"])):
        fmarkers = gen_flag_data_markers(source.data[f"y{yidx}"], fid=fid,
                                         markers=markers, fmarker=fmarker)
        # update the data source with markers
        source.add(fmarkers, name="fmarkers")

        p_glyph = Scatter(x='x', y=f"y{yidx}", marker="fmarkers",
                          line_width=0, angle=0.7, **glyph_opts)
    else:
        p_glyph = Scatter(x='x', y=f"y{yidx}", marker=markers[fid],
                          line_width=0, **glyph_opts)

    # add a check for whether all the in y data were NaNs
    # this causes the errorbars to fail if they all are
    # they must be checked
    if np.all(np.isnan(source.data[f"y{yidx}"])):
        ebars = errorbar(x=source.data['x'], y=source.data[f"y{yidx}"],
                         color=color, yerr=None)
    else:
        ebars = errorbar(x=source.data['x'], y=source.data[f"y{yidx}"],
                         color=color, yerr=yerr)
        p_glyph.js_link("fill_alpha", ebars, "line_alpha")

    logger.debug("Glyphs generated")

    return p_glyph, ebars


################### Stats Related Funcs ################################

def create_stats_table(stats, yaxes):
    """ Create data table with median statistics
    Parameters
    ----------
    stats : :obj:`list`
        List of lists containing data stats for each iterations from :func:`ragavi.ragavi.stats_display`
    yaxes : :obj:`list`
        Contains y-axes for the current plot

    Returns
    -------
        Bokeh column layout containing data table with stats
    """
    # number of y-axes
    n_ys = len(yaxes)
    # number of fields, spws and corrs
    n_items = len(stats) // n_ys
    stats = np.array(stats)
    d_stats = dict(
        spw=stats[:n_items, 0],
        field=stats[:n_items, 1],
        corr=stats[:n_items, 2],
    )

    # get the data in a more useable format
    datas = stats[:, 3].reshape(-1, n_items).T
    for y in range(n_ys):
        d_stats[yaxes[y]] = datas[:, y]

    source = ColumnDataSource(data=d_stats)
    cols = "spw field corr".split() + yaxes

    columns = [TableColumn(field=x, title=x.capitalize()) for x in cols]
    dtab = DataTable(source=source, columns=columns,
                     fit_columns=True, height=150,
                     max_height=180, max_width=600,
                     sizing_mode="stretch_both", width=500)
    t_title = Div(text="Median Statistics")

    logger.debug("Stats table generated")
    return column([t_title, dtab], sizing_mode="stretch_both")


def make_table_name(tab_name):
    """Create div for stats data table"""
    div = Div(text="Table: {}".format(tab_name))
    return div


def stats_display(tab_name, yaxis, corr, field, f_names=None,
                  flag=True, spwid=None):
    """Display some statistics on the plots.
    These statistics are derived from a specific correlation and a specified field of the data.

    Note
    ----
        Currently, only the medians of these plots are displayed.

    Parameters
    ----------
    corr : :obj:`int`
        Correlation number of the data to be displayed
    f_names : :obj:`list`
        List with all the available field names
    field : :obj:`int`
        Integer field id of the field being plotted. If a string name was provided, it will be converted within the main function by
        :meth:`ragavi.vis_utils.name_2id`.
    flag : :obj:`bool`
        Whether to flag data or not
    spwid : :obj:`int`
        Spectral window to be selected
    yaxis : :obj:`str`
        Can be amplitude, phase, real, imaginary or delay

    Returns
    -------
    List containing stats
    """

    logger.debug(vu.ctext("Computing data for stats table"))

    subtable, gtype = get_table(tab_name, fid=field, spwid=str(spwid),
                                group_cols=[], where=[])
    subtable = subtable[0]
    if subtable.row.size == 0:
        return None

    dobj = DataCoreProcessor(subtable, tab_name, gtype, yaxis,
                             corr=corr, flag=True, ddid=spwid)

    y = dobj.y_only(yaxis).y.compute()

    med_y = np.nanmedian(y)

    logger.debug(vu.ctext(f"y-axis: {yaxis}, field: {f_names[field]}, corr: {str(corr)} median: {str(med_y)}"))

    return [spwid, f_names[field], corr, med_y, yaxis]


def gen_checkbox_labels(batch_size, num_leg_objs, antnames):
    """ Auto-generating Check box labels

    Parameters
    ----------
    batch_size : :obj:`int`
        Number of items in a single batch
    num_leg_objs : :obj:`int`
        Number of legend objects / Number of batches
    Returns
    ------
    labels : :obj:`list`
        Batch labels for the batch selection check box group
    """
    nants = len(antnames)

    labels = []
    s = 0
    e = batch_size - 1
    for i in range(num_leg_objs):
        if e < nants:
            labels.append("{} - {}".format(antnames[s], antnames[e]))
        else:
            labels.append("{} - {}".format(antnames[s], antnames[nants - 1]))
        # after each append, move start number to current+batchsize
        s = s + batch_size
        e = e + batch_size

    return labels


################## Saving Plot funcs ###################################

def save_html(name, plot_layout):
    """Save plots in HTML format

    Parameters
    ----------
    hname : :obj:`str`
        HTML Output file name
    plot_layout : :obj:`bokeh.layouts`
                  Layout of the Bokeh plot, could be row, column, gridplot.
    """
    if "html" not in name:
        name = name + ".html"

    logger.debug(f"Saving items to {name}")

    output_file(name)
    output = save(plot_layout, name, title=name)
    logger.info(f"Rendered HTML: {output}")


def save_static_image(name, figs=None):
    """Save plots in SVG format

    Note
    ----
        The number of resulting images depends on `doplot`. By default, this means to objects unless otherwise.
        For SVG, the python package `selenium`, and node package `phantomjs` are required. More information `Exporting bokeh plots <https://bokeh.pydata.org/en/latest/docs/user_guide/export.html>`_

    Parameters
    ----------
    name : :obj:`str`
        Desired image name
    figs : :obj:`list`
         A list containing :obj:`bokeh.plotting.Plot` objects (The figures to be plotted.)

    """
    from selenium import webdriver

    logger.debug("Setting up Firefox selenium driver")
    b_opts = webdriver.FirefoxOptions()
    # Ensure browser is headless
    b_opts.headless = True

    logger.debug(f"Browser headless mode: {b_opts.headless}")

    driver = webdriver.Firefox(options=b_opts)

    logger.debug("Driver successfully set up")

    pren, suffix = name.split('.')

    for _i in range(len(figs)):
        legs = figs[_i].legend

        for _l in legs:
            _l.visible = True

        for _r in figs[_i].renderers:
            _r.visible = True
            _r.glyph.size = 7
            if suffix == "svg":
                if _r.glyph.marker == "circle":
                    _r.glyph.marker = "square_x"
                    _r.glyph.line_width = 1
                    _r.glyph.line_color = "#0f9af0"

        figs[_i].width = 600
        figs[_i].frame_width = int(0.9 * 600)
        figs[_i].height = 500
        figs[_i].frame_height = int(0.9 * 500)

    # some shared options
    o_opts = dict(filename=f"{pren}.{suffix}",
                  timeout=30,
                  webdriver=driver)

    logger.info(f"Attempting to save image in format: {suffix}, name:{pren}")
    if "png" == suffix.lower():
        figs = row(figs)
        store = export_png(obj=figs, **o_opts)

    elif "svg" == suffix.lower():
        # Bokeh SVG doesn't work with circle markers
        # Issue is here: https://github.com/bokeh/bokeh/issues/8446
        logger.info(
            "All circle markers switched to 'square_x' because of SVG issues")
        for fig in figs:
            fig.output_backend = "svg"
        figs = row(figs)
        store = export_svgs(figs, **o_opts)

    logger.info(f"Done. Image at: {store}")

    logger.debug("Attempting to close browser")
    driver.close()


################### Main ###############################################
def main(**kwargs):
    """Main function that launches the gains plotter"""
    if "options" in kwargs:
        NB_RENDER = False
        # capture the parser options
        options = kwargs.get("options", None)

        corr = options.corr
        ddid = options.ddid
        doplot = options.doplot
        fields = options.fields
        html_name = options.html_name
        image_name = options.image_name
        mytabs = options.mytabs
        plotants = options.plotants
        t0 = options.t0
        t1 = options.t1
        where = options.where
        kx = options.kx

    if isinstance(options.fields, list):
        fields = ",".join(fields)

    if options.debug:
        vu.update_log_levels(logger, "debug")
    else:
        vu.update_log_levels(logger, "info")

    if options.logfile:
        vu.update_logfile_name(logger, options.logfile)

    tables = [os.path.abspath(tab) for tab in options.mytabs]

    # parent container of the items in this pot
    final_layout = []

    for tab in tables:

        if options.fields:
            # Not using .isalnum coz the actual field names can be provided
            if (any(_ in fields for _ in ":~") or
                    fields.isnumeric()):
                fields = vu.resolve_ranges(fields)
            elif ',' in fields:
                """
                 convert field name to field id and join all the resulting field ids with a comma
                """
                fields = ",".join([str(vu.name_2id(tab, x))
                                   if not x.isnumeric() else x
                                   for x in fields.split(',')])
                fields = vu.resolve_ranges(fields)
            else:
                fields = str(vu.name_2id(tab, fields))
                fields = vu.resolve_ranges(fields)

        if options.corr:
            if ',' in options.corr:
                corrs = [int(_) for _ in options.corr.split(',')]
            else:
                corrs = [int(options.corr)]
            corrs = np.array(corrs)
        else:
            s_tab = table(tab, ack=False)
            corrs = np.arange(s_tab.getcell("FLAG", 0).shape[-1])
            s_tab.close()

        if options.ddid is not None:
            # TODO: Check for the input coming in from argparse
            # check if it needs to be resolved
            ddid = vu.resolve_ranges(options.ddid)
            # get all frequencies anyways
            freqs = vu.get_frequencies(tab).values / _GHZ_
        else:
            freqs = vu.get_frequencies(tab).values / _GHZ_

        if options.plotants:
            plotants = vu.resolve_ranges(options.plotants)

        init_time, end_time = get_time_range(tab, unix_time=False)

        if options.where is None:
            where = []
        else:
            where = [where]

        if options.mycmap is None:
            cmap = "coolwarm"
        else:
            cmap = options.mycmap

        # perform a time selections
        if t0:
            where.append(f"TIME - {str(init_time)} >= {str(t0)}")
        if t1:
            where.append(f"TIME - {str(init_time)} <= {str(t1)}")

        all_figures = []

        # store generated plots
        all_glyphs = []

        # store legend items
        all_legends = []

        all_ebars = []
        # ebars_ax2 = []

        # Store stats
        all_stats = []

        # store flagged and unflaged data stources
        all_ufsources = []
        all_fsources = []

        subs, gain = get_table(tab, spwid=ddid, where=where, fid=fields,
                               antenna=plotants)

        if doplot == "ap":
            y_axes = ["amplitude", "phase"]
        else:
            y_axes = ["real", "imaginary"]
        if gain == "K":
            y_axes = ["delay"]

        if gain in ["G", "F"] or (options.kx == "time" and gain == "K"):
            x_axis_type = "datetime"
        else:
            x_axis_type = "linear"

        # confirm a populous table is selected
        try:
            assert len(subs) > 0
        except AssertionError:
            logger.info(
                "Table contains no data. Check data selection. Skipping.")
            continue

        spw_ids = np.unique(np.array([s.SPECTRAL_WINDOW_ID for s in subs]))

        # get antenna names
        ant_names = vu.get_antennas(tab).values
        ant_ids = np.unique(np.array([s.ANTENNA1 for s in subs]))

        field_names = vu.get_fields(tab).values
        field_ids = np.unique(np.array([s.FIELD_ID for s in subs]))

        field_ucodes = [u"\u2B24", u"\u25C6", u"\u25FC", u"\u25B2",
                        u"\u25BC", u"\u2B22"]

        # setting up colors for the antenna plot
        cmap = vu.get_linear_cmap(cmap, ant_ids.size)

        for _y, yaxis in enumerate(y_axes, start=1):
            fig_glyphs = []
            fig_renderers = []
            fig_legends = []
            fig_ebars = []
            fig_stats = []
            ufsources = []
            fsources = []

            iters = product(spw_ids, field_ids, corrs)

            for spw, fid, corr in iters:
                fname = field_names[fid]

                logger.info(f"Spw: {spw}, Field: {fname}, Corr: {corr} {yaxis}")
                stats = stats_display(tab_name=tab, yaxis=yaxis,
                                      corr=corr, field=fid, flag=_FLAG_DATA_,
                                      f_names=field_names, spwid=spw)
                if stats:
                    fig_stats.append(stats)

                for _a, ant in enumerate(ant_ids):
                    legend = ant_names[ant]
                    colour = cmap[_a]
                    for sub in subs:
                        if (sub.SPECTRAL_WINDOW_ID == spw and
                                sub.FIELD_ID == fid and
                                sub.ANTENNA1 == ant):
                            # for tooltips
                            spw_id, scan = get_tooltip_data(sub, gain, freqs)
                            source = ColumnDataSource(
                                data={"scanid": scan,
                                      "corr":  np.full(scan.shape, corr,
                                                       dtype=np.uint8),
                                      "field": np.full(scan.shape, fname),
                                      "spw": spw_id,
                                      "antname": np.full(scan.shape, legend)
                                      })
                            inv_source = ColumnDataSource(data={})

                            logger.debug(vu.ctext(f"Processing un-flagged data for {legend}"))

                            data_obj = DataCoreProcessor(
                                sub, tab, gain,
                                fid=fid, yaxis=yaxis, corr=corr,
                                flag=_FLAG_DATA_, kx=options.kx, ddid=spw)

                            data = data_obj.act()
                            xaxis = data_obj.xaxis

                            logger.debug(f"Processing flagged data")
                            infl_data = DataCoreProcessor(
                                sub, tab, gain, fid=fid, yaxis=yaxis,
                                corr=corr, flag=not _FLAG_DATA_,
                                kx=options.kx, ddid=spw)

                            infl_data = infl_data.act()

                            x = data.x
                            x_label = data.x_label

                            y = data.y
                            y_err = data.y_err
                            y_label = data.y_label

                            iy = infl_data.y

                            source.add(x, name='x')
                            source.add(y, name=f"y{_y}")

                            inv_source.add(y, name=f"y{_y}")
                            inv_source.add(iy, name=f"iy{_y}")

                            ufsources.append(source)
                            fsources.append(inv_source)

                            glyphs, bars = make_plots(
                                source=source,
                                color=colour,
                                fid=np.where(field_ids == fid)[0][0],
                                yerr=y_err,
                                yidx=_y)

                            fig_glyphs.append(glyphs)
                            fig_legends.append((legend, []))
                            fig_ebars.append(bars)
                            sub.close()

            title = f"{yaxis.capitalize()} vs {xaxis.capitalize()}"
            fig = create_bk_fig(xlab=x_label, ylab=y_label, title=title,
                                x_min=x.min(), x_max=x.max(),
                                x_axis_type=x_axis_type, x_name=xaxis,
                                y_name=yaxis, pw=_PLOT_WIDTH_,
                                ph=_PLOT_HEIGHT_, add_grid=True,
                                add_title=True, add_xaxis=True,
                                add_yaxis=True, fix_plotsize=False,
                                ti_specs=dict(desired_num_ticks=8),
                                ax_specs=dict(minor_tick_line_alpha=1))

            h_tool = fig.select(name="p_htool")[0]

            tooltips = [("({:.4}, {:.4})".format(x_label, y_label),
                         f"(@x, @y{_y})"),
                        ("spw", "@spw"), ("scan_id", "@scanid"),
                        ("antenna", "@antname"),
                        ("corr", "@corr"), ("field", "@field")]

            if xaxis == "channel":
                for spw in spw_ids:
                    # do this here otherwise extra x added after title
                    title = fig.above.pop()
                    logger.debug("Adding extra chan/freq axis")
                    fig = add_axis(fig, (freqs[spw][0], freqs[spw][-1]),
                                   ax_label="Frequency [GHz]",
                                   ax_name=f"spw{spw}")
                    fig.add_layout(title, "above")
            elif xaxis == "time":
                munix = 3506716800.0
                ni_time = datetime.utcfromtimestamp(
                    init_time - munix).strftime("%d %b %Y, %H:%M:%S")
                ne_time = datetime.utcfromtimestamp(
                    end_time - munix).strftime("%d %b %Y, %H:%M:%S")

                fig.xaxis.axis_label = "{} from {} UTC".format(
                    x_label, ni_time)

                # format tootltip to time value
                h_tool.formatters = {"@x": "datetime"}

                tooltips[0] = ("({:.4}, {:.4})".format(x_label, y_label),
                               "(@x{%F %T}, $y)")

            if yaxis == "phase":
                # Add degree sign to phase tick formatter
                fig.yaxis.formatter = PrintfTickFormatter(
                    format=u"%f\u00b0")

            h_tool.tooltips = tooltips

            n_cds = len(ufsources)

            for _s in range(n_cds):
                p = fig.add_glyph(ufsources[_s], fig_glyphs[_s])

                if _s >= _BATCH_SIZE_:
                    p.visible = False

                fig_renderers.append(p)
                # form legend item
                fig_legends[_s][1].append(p)
                # add error bars
                if fig_ebars[_s]:
                    fig.add_layout(fig_ebars[_s])

            n_leg_objs = int(np.ceil(len(ant_ids) / _BATCH_SIZE_))

            leg_batches = create_legend_batches(n_leg_objs, fig_legends,
                                                batch_size=_BATCH_SIZE_)

            leg_objs = create_legend_objs(n_leg_objs, leg_batches)

            # make the last created legend object the bottom one
            for i in reversed(range(n_leg_objs)):
                fig.add_layout(leg_objs[f"leg_{str(i)}"], "above")
                all_legends.append(leg_objs[f"leg_{str(i)}"])

            all_glyphs.append(fig_glyphs)
            all_ebars.append(fig_ebars)
            all_fsources.append(fsources)
            all_ufsources.append(ufsources)
            all_stats.extend(fig_stats)
            all_figures.append(fig)

        stats_table = create_stats_table(all_stats, y_axes)

        # linking a bunch of stuff
        if len(all_figures) > 1:
            all_fsources, all_ufsources = link_plots(
                all_figures=all_figures, all_fsources=all_fsources,
                all_ebars=all_ebars)

        ######################################################################
        ################ Defining widgets ###################################
        ######################################################################

        logger.debug("Defining plot widgets")

        # widget dimensions
        w_dims = dict(width=150, height=30)

        # Selection group
        # creating and configuring Antenna selection buttons
        ant_select = CheckboxGroup(labels=["Select all antennas"],
                                   active=[], **w_dims)

        ant_labs = gen_checkbox_labels(_BATCH_SIZE_, n_leg_objs, ant_names)
        batch_select = CheckboxGroup(labels=ant_labs, active=[0],
                                     **w_dims)

        corr_labs = ["Correlation {}".format(str(_)) for _ in corrs]
        corr_select = CheckboxGroup(labels=corr_labs, active=[0],
                                    width=150)

        field_labels = ["Field {} {}".format(
            field_names[int(x)],
            field_ucodes[enum_fid]) for enum_fid, x in enumerate(field_ids)]

        field_selector = CheckboxGroup(labels=field_labels, active=[0],
                                       **w_dims)

        spw_labs = ["Spw: {}".format(str(_)) for _ in spw_ids]
        spw_select = CheckboxGroup(labels=spw_labs, active=[0], width=150)

        # Additions group
        # Checkbox to hide and show legends
        legend_toggle = CheckboxGroup(labels=["Show legends"], active=[],
                                      **w_dims)

        save_selected = Button(label="Download data selection",
                               button_type="success", margin=(7, 5, 3, 5),
                               **w_dims)

        # configuring toggle button for showing all the errors
        toggle_err = CheckboxGroup(labels=["Show error bars"], active=[],
                                   **w_dims)

        toggle_flag = CheckboxGroup(labels=["Show flagged data"],
                                    active=[], **w_dims)

        # creating glyph size slider for the plots
        # margin = [top, right, bottom, left]
        size_slider = Slider(end=15, start=0.4, step=0.1,
                             value=4, title="Glyph size", margin=(3, 5, 7, 5),
                             bar_color="#6F95C3", ** w_dims)

        # Alpha slider for the glyphs
        alpha_slider = Slider(end=1, start=0.1, step=0.1, value=1,
                              margin=(3, 5, 7, 5), title="Glyph alpha",
                              bar_color="#6F95C3", **w_dims)

        # Plot layout options
        axis_fontslider = Slider(end=20, start=3, step=0.5, value=10,
                                 margin=(7, 5, 3, 5), title="Axis label size",
                                 bar_color="#6F95C3", **w_dims)
        title_fontslider = Slider(end=35, start=10, step=1, value=15,
                                  margin=(3, 5, 7, 5), title="Title size",
                                  bar_color="#6F95C3", **w_dims)

        tname_div = make_table_name(tab)

        ######################################################################
        ############## Defining widget Callbacks ############################
        ######################################################################

        logger.debug("Attaching callbacks to widgets")

        ant_select.js_on_change("active", CustomJS(
            args=dict(ax=all_figures[0].renderers, bsel=batch_select,
                      fsel=field_selector, csel=corr_select, ssel=spw_select,
                      nbatches=n_leg_objs, nfields=field_ids.size,
                      ncorrs=corrs.size, nspws=spw_ids.size),
            code=ant_select_callback()))

        # BATCH SELECTION
        batch_select.js_on_change("active", CustomJS(
            args=dict(ax=all_figures[0].renderers, bsize=_BATCH_SIZE_,
                      nants=ant_ids.size, nbatches=n_leg_objs,
                      nfields=field_ids.size, ncorrs=corrs.size,
                      nspws=spw_ids.size, fsel=field_selector,
                      csel=corr_select, antsel=ant_select, ssel=spw_select),
            code=batch_select_callback()))

        corr_select.js_on_change("active", CustomJS(
            args=dict(bsel=batch_select, bsize=_BATCH_SIZE_,
                      fsel=field_selector, nants=ant_ids.size,
                      ncorrs=corrs.size, nfields=field_ids.size,
                      nbatches=n_leg_objs, nspws=spw_ids.size,
                      ax=all_figures[0].renderers, ssel=spw_select),
            code=corr_select_callback()))

        field_selector.js_on_change("active", CustomJS(
            args=dict(bsize=_BATCH_SIZE_, bsel=batch_select, csel=corr_select,
                      nants=ant_ids.size, nfields=field_ids.size,
                      ncorrs=corrs.size, nbatches=n_leg_objs,
                      nspws=spw_ids.size, ax=all_figures[0].renderers,
                      ssel=spw_select),
            code=field_selector_callback()))

        ex_ax = all_figures[0].select(type=LinearAxis, layout="above")
        ex_ax = sorted({_.id: _ for _ in ex_ax}.items())
        ex_ax = [_[1] for _ in ex_ax]

        spw_select.js_on_change("active", CustomJS(
            args=dict(bsel=batch_select, bsize=_BATCH_SIZE_, csel=corr_select,
                      fsel=field_selector, nants=ant_ids.size,
                      ncorrs=corrs.size, nfields=field_ids.size,
                      nbatches=n_leg_objs, nspws=spw_ids.size,
                      ax=all_figures[0].renderers, spw_ids=spw_ids,
                      ex_ax=ex_ax),
            code=spw_select_callback()))

        legend_toggle.js_on_change("active", CustomJS(
            args=dict(legs=all_legends, n_legs=n_leg_objs),
            code=legend_toggle_callback()))

        toggle_err.js_on_change("active", CustomJS(
            args=dict(ax=all_figures[0].renderers,
                      ax_err=all_ebars[0]),
            code=toggle_err_callback()))

        toggle_flag.js_on_change("active", CustomJS(
            args=dict(f_sources=all_fsources,
                      uf_sources=all_ufsources),
            code=flag_callback()))

        save_selected.js_on_click(CustomJS(args=dict(
            uf_src=all_ufsources,
            f_src=all_fsources),
            code=save_selected_callback()))

        alpha_slider.js_on_change("value",
                                  CustomJS(args=dict(
                                      ax=all_glyphs[0]),
                                      code=alpha_slider_callback()))

        size_slider.js_on_change("value",
                                 CustomJS(args=dict(slide=size_slider,
                                                    ax=all_glyphs[0]),
                                          code=size_slider_callback()))

        axis_fontslider.js_on_change("value",
                                     CustomJS(args=dict(
                                         ax=all_figures[0].axis),
                                         code=axis_fs_callback()))
        title_fontslider.js_on_change(
            "value",
            CustomJS(args=dict(ax=all_figures[0].select(name="p_title")[0]),
                     code=title_fs_callback()))

        #################################################################
        ########## Define widget layouts #################################
        ##################################################################

        logger.debug("Setting up plot layout")

        asel_div = Div(text="Select antenna group")
        fsel_div = Div(text="Fields")
        ssel_div = Div(text="Select spw")
        csel_div = Div(text="Select correlation")

        w_box = grid(
            children=[
                [ant_select, toggle_err, size_slider, title_fontslider],
                [legend_toggle, toggle_flag, alpha_slider, axis_fontslider],
                [asel_div, fsel_div, ssel_div, csel_div],
                [batch_select, field_selector, spw_select, corr_select]
            ])

        all_widgets = row([w_box, save_selected, stats_table],
                          sizing_mode=None, spacing=10, max_height=180)

        plots = gridplot([all_figures], toolbar_location="above",
                         sizing_mode="stretch_width")

        lay = layout([[tname_div], [all_widgets], [plots]],
                     sizing_mode="stretch_width")
        final_layout.append(lay)

        logger.info("Table {} done.".format(tab))

    if _NB_RENDER_:
        return final_layout
    else:
        if options.image_name and html_name:
            save_html(html_name, final_layout)
            save_static_image(name=options.image_name, figs=all_figures)

        elif options.image_name:
            save_static_image(name=options.image_name, figs=all_figures)
        elif html_name:
            save_html(html_name, final_layout)
        else:
            t_name = os.path.basename(tables[0])
            html_name = f"{t_name}_{doplot}"
            if len(tables) > 1:
                t_now = datetime.now().strftime("%Y%m%d_%H%M%S")
                html_name = html_name.replace(t_name, t_now)
            save_html(html_name, final_layout)
        return 0


##################### Notebook Entry point #############################
def plot_table(**kwargs):
    """Plot gain tables within Jupyter notebooks. Parameter names correspond to the long names of each argument (i.e those with --) from the `ragavi-vis` command line help

    Parameters
    ----------
    table : :obj:`str` or :obj:`list`
        The table (list of tables) to be plotted.

    ant : :obj:`str, optional`
        Plot only specific antennas, or comma-separated list of antennas.
    corr : :obj:`int, optional`
        Correlation index to plot. Can be a single integer or comma separated
        integers e.g "0,2". Defaults to all.
    cmap : `str, optional`
        Matplotlib colour map to use for antennas. Default is coolwarm
    ddid : :obj:`int`
        SPECTRAL_WINDOW_ID or ddid number. Defaults to all
    doplot : :obj:`str, optional`
        Plot complex values as amp and phase (ap) or real and imag (ri).
        Default is "ap".
    field : :obj:`str, optional`
        Field ID(s) / NAME(s) to plot. Can be specified as "0", "0,2,4",
        "0~3" (inclusive range), "0:3" (exclusive range), "3:" (from 3 to
        last) or using a field name or comma separated field names. Defaults
        to all.
    k-xaxis: :obj:`str`
        Choose the x-xaxis for the K table. Valid choices are: time or
        antenna. Defaults to time.
    taql: :obj:`str, optional`
        TAQL where clause
    t0  : :obj:`int, optional`
        Minimum time [in seconds] to plot. Default is full range
    t1 : :obj:`int, optional`
        Maximum time [in seconds] to plot. Default is full range

    """
    from ragavi.arguments import gains_argparser

    nargs = []
    for k, v in kwargs.items():
        nargs.append(f"--{k}")
        if isinstance(v, list):
            nargs.extend(v)
        else:
            nargs.append(v)
    parser = gains_argparser()
    options = parser.parse_args(nargs)

    # set NB rendering to true
    global _NB_RENDER_
    _NB_RENDER_ = True
    output_notebook()

    lays = main(options=options)

    for _l in lays:
        show(_l)
    logger.info("Done Plotting")

    return



# -*- coding: utf-8 -*-

"""
This is a module to average the main table of Measurement Sets, as well as the spectral window table. It largely borrows and was adapted from the xova application averager, whose code can be found here:
https://github.com/ska-sa/xova/blob/22614b64f0dd8caca20db74ec29705711e347fd0/xova/apps/xova/averaging.py
Commit #d653b38

"""

import logging
import dask.array as da
import daskms as xm
import numpy as np
import xarray as xr

from collections import OrderedDict
from dask.array.reductions import partial_reduce
from daskms import Dataset
from itertools import product, combinations
from africanus.averaging.dask import (time_and_channel,
                                      chan_metadata,
                                      chan_average as dask_chan_avg)

import ragavi.utils as vu

logger = logging.getLogger(__name__)


def _id(array, fill_value=0, dtype_=np.int32):
    return np.full_like(array, fill_value, dtype=dtype_)


def id_full_like(exemplar, fill_value, dtype=np.int32):
    """ full_like that handles nan chunk sizes """
    return exemplar.map_blocks(_id, fill_value=fill_value,
                               dtype_=dtype, dtype=dtype)


def _safe_concatenate(*args):
    # Handle list with single numpy array case,
    # tuple unpacking fails on it
    if len(args) == 1 and isinstance(args[0], np.ndarray):
        return args[0]

    return np.concatenate(*args)


def concatenate_row_chunks(array, group_every=4):
    """
    Parameters
    ----------
    array : :class:`dask.array.Array`
        dask array to average.
        First dimension must correspond to the MS 'row' dimension
    group_every : int
        Number of adjust dask array chunks to group together.
        Defaults to 4.
    When averaging, the output array's are substantially smaller, which
    can affect disk I/O since many small operations are submitted.
    This operation concatenates row chunks together so that more rows
    are submitted at once
    """

    # Single chunk already
    if len(array.chunks[0]) == 1:
        return array

    # Restrict the number of chunks to group to the
    # actual number of chunks in the array
    group_every = min(len(array.chunks[0]), group_every)
    data = partial_reduce(_safe_concatenate, array,
                          split_every={0: group_every},
                          reduced_meta=None, keepdims=True)

    # partial_reduce sets the number of rows in each chunk
    # to 1, which is untrue. Correctly set the row chunks to nan,
    # steal the graph and recreate the array
    row_chunks = tuple(np.nan for _ in data.chunks[0])
    chunks = (row_chunks,) + data.chunks[1:]
    graph = data.__dask_graph__()

    return da.Array(graph, data.name, chunks, dtype=data.dtype)


def output_dataset(avg, field_id, data_desc_id, scan_number,
                   group_row_chunks, viscolumn=""):
    """
    Parameters
    ----------
    avg : namedtuple
        Result of :func:`average`
    field_id : int
        FIELD_ID for this averaged data
    data_desc_id : int
        DATA_DESC_ID for this averaged data
    scan_number : int
        SCAN_NUMBER for this averaged data
    group_row_chunks : int
        Concatenate row chunks group

    Returns
    -------
    Dataset
        Dataset containing averaged data
    """
    # Create ID columns
    fid = field_id
    ddid = data_desc_id
    scn = scan_number
    field_id = id_full_like(avg.time, fill_value=field_id)
    data_desc_id = id_full_like(avg.time, fill_value=data_desc_id)
    scan_number = id_full_like(avg.time, fill_value=scan_number)

    # Single flag category, equal to flags
    flag_cats = avg.flag[:, None, :, :]

    out_ds = {
        "ANTENNA1": (("row",), avg.antenna1),
        "ANTENNA2": (("row",), avg.antenna2),
        "DATA_DESC_ID": (("row",), data_desc_id),
        "FIELD_ID": (("row",), field_id),
        "SCAN_NUMBER": (("row",), scan_number),
        "FLAG_ROW": (("row",), avg.flag_row),
        # "FLAG_CATEGORY": (("row", "flagcat", "chan", "corr"), flag_cats),
        "TIME": (("row",), avg.time),
        "INTERVAL": (("row",), avg.interval),
        "TIME_CENTROID": (("row",), avg.time_centroid),
        "EXPOSURE": (("row",), avg.exposure),
        "UVW": (("row", "uvw"), avg.uvw),
        "WEIGHT": (("row", "corr"), avg.weight),
        "SIGMA": (("row", "corr"), avg.sigma),
        viscolumn: (("row", "chan", "corr"), avg.vis),
        "FLAG": (("row", "chan", "corr"), avg.flag),
    }

    # Add optionally averaged columns columns
    if avg.weight_spectrum is not None:
        out_ds["WEIGHT_SPECTRUM"] = (("row", "chan", "corr"),
                                     avg.weight_spectrum)

    if avg.sigma_spectrum is not None:
        out_ds["SIGMA_SPECTRUM"] = (("row", "chan", "corr"),
                                    avg.sigma_spectrum)

    # Concatenate row chunks together
    if group_row_chunks > 1:
        grc = group_row_chunks
        # Remove items whose values are None
        out_ds = {k: (dims, concatenate_row_chunks(data, group_every=grc))
                  for k, (dims, data) in out_ds.items() if data is not None}

    return Dataset(out_ds, attrs={"DATA_DESC_ID": ddid,
                                  "FIELD_ID": fid,
                                  "SCAN_NUMBER": scn})


def average_main(main_ds, time_bin_secs, chan_bin_size,
                 group_row_chunks, respect_flag_row=True,
                 viscolumn="DATA", sel_cols=None):
    """ Perform averaging of the main table of the MS
    At this point,the input :attr:`main_ds` is a list containing an MS
    grouped by DDID, FIELD_ID & SCAN_NUMBER.

    Parameters
    ----------
    main_ds : list of Datasets
        Dataset containing Measurement Set columns.
        Should have a DATA_DESC_ID attribute.
    time_bin_secs : float
        Number of time bins to average together
    chan_bin_size : int
        Number of channels to average together
    group_row_chunks : int, optional
        Number of row chunks to concatenate together
    respect_flag_row : bool
        Respect FLAG_ROW instead of using FLAG
        for computing row flags.
    viscolumn: string
        name of column to average
    sel_cols: list
        Columns that need to be present in the dataset
    Returns
    -------
    avg
        tuple containing averaged data
    """
    output_ds = []

    # for each group
    for ds in main_ds:
        if respect_flag_row is False:
            if len(ds.FLAG.dims) > 2:
                ds = ds.assign(
                    FLAG_ROW=(("row",),
                              ds.FLAG.data.all(axis=(1, 2))))
            else:
                ds = ds.assign(FLAG_ROW=(("row",),
                                         ds.FLAG.data.all(axis=1)))

        # store the subgroup's data variables
        dv = ds.data_vars

        # Default kwargs.
        kwargs = {"time_bin_secs": time_bin_secs,
                  "chan_bin_size": chan_bin_size,
                  "vis": dv[viscolumn].data}

        # Other columns with directly transferable names
        # columns = ["FLAG_ROW", "TIME_CENTROID", "EXPOSURE", "WEIGHT",
        #           "SIGMA",
        #            "UVW", "FLAG", "WEIGHT_SPECTRUM", "SIGMA_SPECTRUM"]
        columns = ["TIME_CENTROID"] + sel_cols

        for c in columns:
            if c != viscolumn:
                try:
                    kwargs[c.lower()] = dv[c].data
                except KeyError:
                    pass

        # Set up the average operation
        avg = time_and_channel(**kwargs)

        output_ds.append(output_dataset(avg,
                                        ds.FIELD_ID,
                                        ds.DATA_DESC_ID,
                                        ds.SCAN_NUMBER,
                                        group_row_chunks,
                                        viscolumn=viscolumn))

    return output_ds


def average_spw(spw_ds, chan_bin_size):
    """
    Parameters
    ----------
    spw_ds : list of Datasets
        list of Datasets, each describing a single Spectral Window
    chan_bin_size : int
        Number of channels in an averaging bin
    Returns
    -------
    spw_ds : list of Datasets
        list of Datasets, each describing an averaged Spectral Window
    """

    new_spw_ds = []

    for r, spw in enumerate(spw_ds):
        # Get the dataset variables as a mutable dictionary
        dv = dict(spw.data_vars)

        # Extract arrays we wish to average
        chan_freq = dv["CHAN_FREQ"].data[0]
        chan_width = dv["CHAN_WIDTH"].data[0]
        effective_bw = dv["EFFECTIVE_BW"].data[0]
        resolution = dv["RESOLUTION"].data[0]

        # Construct channel metadata
        chan_arrays = (chan_freq, chan_width, effective_bw, resolution)
        chan_meta = chan_metadata((), chan_arrays, chan_bin_size)
        # Average channel based data
        avg = dask_chan_avg(chan_meta, chan_freq=chan_freq,
                            chan_width=chan_width,
                            effective_bw=effective_bw,
                            resolution=resolution,
                            chan_bin_size=chan_bin_size)

        num_chan = da.full((1,), avg.chan_freq.shape[0], dtype=np.int32)

        # These columns change, re-create them
        dv["NUM_CHAN"] = (("row",), num_chan)
        dv["CHAN_FREQ"] = (("row", "chan"), avg.chan_freq[None, :])
        dv["CHAN_WIDTH"] = (("row", "chan"), avg.chan_width[None, :])
        dv["EFFECTIVE_BW"] = (("row", "chan"), avg.effective_bw[None, :])
        dv["RESOLUTION"] = (("row", "chan"), avg.resolution[None, :])

        # But re-use all the others
        new_spw_ds.append(Dataset(dv))

    return new_spw_ds


def get_averaged_ms(ms_name, tbin=None, cbin=None, chunks=None, taql_where='',
                    columns=None, chan=None, corr=None, data_col=None,
                    group_cols=None, iter_axis=None):
    """ Performs MS averaging.

    Before averaging is performed, data selections is already performed 
    during the MS  acquisition process. TAQL (if available) is used to 
    perform selections for FIELD, SPW/DDID & SCAN. This is the first round of 
    selection. The second round involves selections over channels and 
    correlations. This is done via a slicer. With the exception of corr 
    selectino, all the other selections are done before averaging. This is 
    done because the averager requires 3-dimensional data.

    MS is then grouped by DDID, FIELD_ID & SCAN_NUMBER and fed into 
    :meth:`average_main` which actually performs the averaging.

    This function returns to  :meth:`ragavi.visibilities.get_ms` and is therefore grouped and column select similarly


    Parameters
    ----------
    ms_name : :obj:`str`
        Name of the input MS
    tbin : :obj:`float`
        Time bin in seconds
    cbin : :obj:`int`
        Number of channels to bin together
    chunks : :obj:`dict`
        Size of resulting MS chunks.
    taql_where: :obj:`str`
        TAQL clause to pass to xarrayms
    columns: :obj:`list`
        Columns to be present in the data
    chan : :obj:`slicer`
        A slicer to select the channels to be present in the dataset
    corr : :obj:`slicer` or :obj:`int`
        Correlation index of slice to be present in the dataset
    data_col : :obj:`str`
        Column containing data to be used
    group_cols: :obj:`list`
        List containing columns by which to group the data
    iter_axis: :obj:`str`
        Axis over which iteration is done

    Returns
    -------
    x_dataset: :obj:`list`
        List of :obj:`xarray.Dataset` containing averaged MS. The MSs are split by Spectral windows and grouped depending on the type of plots.

    """

    if chunks is None:
        chunks = dict(row=10000)

    # these are the defaults in averager
    if tbin is None:
        tbin = 1
    if cbin is None:
        cbin = 1

    # ensure that these are in the selected columns
    for _ in ["TIME", "ANTENNA1", "ANTENNA2", "INTERVAL", "FLAG", "FLAG_ROW", data_col]:
        if _ not in columns:
            columns.append(_)

    # must be grouped this way because of time averaging
    ms_obj = xm.xds_from_ms(ms_name,
                            group_cols=["DATA_DESC_ID",
                                        "FIELD_ID", "SCAN_NUMBER"],
                            columns=columns,
                            taql_where=taql_where)

    # some channels have been selected
    # corr selection is performed after averaging!!
    if chan is not None:
        ms_obj = [_.sel(chan=chan) for _ in ms_obj]

    logger.info("Averaging MAIN table")

    # perform averaging to the MS
    avg_mss = average_main(main_ds=ms_obj, time_bin_secs=tbin,
                           chan_bin_size=cbin, group_row_chunks=100000,
                           respect_flag_row=False, sel_cols=columns,
                           viscolumn=data_col)
    n_ams = len(avg_mss)

    # writes_ms = xm.xds_to_table(avg_mss, "tesxt", "ALL")
    logger.info("Creating averaged xarray Dataset")

    x_datasets = []
    for _a, ams in enumerate(avg_mss):
        ams = ams.compute()
        logger.info(f"Averaging {_a+1} / {n_ams}")

        datas = {k: (v.dims, v.data, v.attrs)
                 for k, v in ams.data_vars.items() if k != "FLAG_CATEGORY"}

        new_ds = xr.Dataset(datas, attrs=ams.attrs, coords=ams.coords)
        new_ds = new_ds.chunk(chunks)

        x_datasets.append(new_ds)

    # data will always be grouped by SPW unless iterating over antenna
    # the most amount of grouping that will occur will be between to columns
    all_grps = []

    if len(group_cols) == 0:
        # return a single dataset
        subs = xr.combine_nested(x_datasets, concat_dim="row",
                                 compat="no_conflicts", data_vars="all",
                                 coords="different", join="outer")
        subs.attrs = {}
        subs = subs.chunk(chunks)
        all_grps.append(subs)

    elif (set(group_cols) <= {"DATA_DESC_ID", "ANTENNA1", "ANTENNA2"} or
          iter_axis == "antenna"):
        uniques = np.unique([_.attrs["DATA_DESC_ID"] for _ in x_datasets])
        uants = np.arange(vu.get_antennas(ms_name).size)

        for _d in uniques:
            subs = []
            for _ in x_datasets:
                if _.attrs["DATA_DESC_ID"] == _d:
                    subs.append(_)
            subs = xr.combine_nested(subs, concat_dim="row",
                                     compat="no_conflicts", data_vars="all",
                                     coords="different", join="outer")
            subs.attrs = {"DATA_DESC_ID": _d}
            subs = subs.chunk(chunks)

            if {"ANTENNA1", "ANTENNA2"} <= set(group_cols):
                u_bl = combinations(uants, 2)
                for p, q in u_bl:
                    n_subs = subs.where((subs.ANTENNA1 == p) &
                                        (subs.ANTENNA2 == q), drop=True)
                    n_subs.attrs = {"DATA_DESC_ID": _d,
                                    "ANTENNA1": p,
                                    "ANTENNA2": q}
                    all_grps.append(n_subs)
            elif "ANTENNA1" in group_cols:
                for p in uants[:-1]:
                    n_subs = subs.where((subs.ANTENNA1 == p), drop=True)
                    n_subs.attrs = {"DATA_DESC_ID": _d,
                                    "ANTENNA1": p}
                    all_grps.append(n_subs)
            elif "ANTENNA2" in group_cols:
                for q in uants[:-1] + 1:
                    n_subs = subs.where((subs.ANTENNA2 == q), drop=True)
                    n_subs.attrs = {"DATA_DESC_ID": _d,
                                    "ANTENNA2": q}
                    all_grps.append(n_subs)
            elif iter_axis == "antenna":
                for p in uants:
                    n_subs = subs.where((subs.ANTENNA1 == p) |
                                        (subs.ANTENNA2 == p), drop=True)
                    n_subs.attrs = {"DATA_DESC_ID": _d,
                                    "ANTENNA": p}
                    all_grps.append(n_subs)
            else:
                all_grps.append(subs)

    elif set(group_cols) <= {"DATA_DESC_ID", "FIELD_ID", "SCAN_NUMBER"}:
        grps = {}
        # must be ddid + something else
        # if it is something other than fid and scan e.g
        # by default group by ddid
        for grp in group_cols:
            uniques = np.unique([_.attrs[grp] for _ in x_datasets])
            grps[grp] = uniques
            # grps.append(uniques)
        for com in product(*grps.values()):
            subs = []
            natt = {k: v for k, v in zip(group_cols, com)}
            for _ in x_datasets:
                if set(natt.items()) <= set(_.attrs.items()):
                    subs.append(_)
            subs = xr.combine_nested(subs, concat_dim="row",
                                     compat="no_conflicts",
                                     data_vars="all", coords="different",
                                     join="outer")
            subs.attrs = natt
            subs = subs.chunk(chunks)
            all_grps.append(subs)

     # select a corr
    if corr is not None:
        all_grps = [_.sel(corr=corr) for _ in all_grps]

    logger.info("Averaging completed.")

    return all_grps


def get_averaged_spws(ms_name, cbin, chan_select=None):
    """ Average spectral windows

    Parameters
    ----------
    ms_name : :obj:`str`
        Path or name of MS Spectral window Subtable
    cbin : :obj:`int`
        Number of channels to be binned together
    chan_select : :obj:`slicer` or :obj:`int`
        Which channels to select 

    Returns
    -------
    x_datasets : :obj:`list`
        List containing averaged spectral windows as :obj:`xarray.Dataset`
    """
    spw_subtab = list(xm.xds_from_table(ms_name, group_cols='__row__'))

    if chan_select is not None:
        spw_subtab = [_.sel(chan=chan_select) for _ in spw_subtab]

    logger.info("Averaging SPECTRAL_WINDOW subtable")

    av_spw = average_spw(spw_subtab, cbin)

    x_datasets = []

    # convert from daskms datasets to xarray datasets
    for sub_ds in av_spw:
        sub_ds = sub_ds.compute()

        datas = {k: (v.dims, v.data, v.attrs)
                 for k, v in sub_ds.data_vars.items() if k != "FLAG_CATEGORY"}

        new_ds = xr.Dataset(datas, attrs=sub_ds.attrs, coords=sub_ds.coords)
        x_datasets.append(new_ds)

    logger.info("Done")
    return x_datasets



import logging

from bokeh.models import (BasicTicker, DatetimeAxis, DataRange1d, Grid,
                          Legend, LinearAxis, LinearScale, LogAxis, LogScale,
                          Toolbar, Plot, PrintfTickFormatter, Range1d,
                          Scatter, Text, Title, Whisker)
from bokeh.models.tools import (BoxEditTool, BoxSelectTool, BoxZoomTool,
                                EditTool, HoverTool, LassoSelectTool, PanTool,
                                ResetTool, RangeTool, SaveTool, UndoTool,
                                WheelZoomTool)

logger = logging.getLogger(__name__)


def create_bk_fig(x=None, xlab=None, x_min=None, x_max=None,
                  ylab=None, fh=None, fw=None,
                  title=None, pw=None, ph=None, x_axis_type="linear",
                  y_axis_type="linear", x_name=None, y_name=None, **kwargs):
    """ Generates a bokeh figure

    Parameters
    ----------
    x :obj:`DataArray`
        Contains x-axis data
    xlab : :obj:`str`
        X-axis label
    x_min : :obj:`float`
        Min x value
    x_max : :obj:`float`
        Max x value
    ylab : :obj:`str`
        Y-axis label
    fh: :obj:`int`
        True height of figure without legends, axes titles etc
    fw: :obj:`int`
        True width of figure without legends, axes etc
    title: :obj:`str`
        Title of plot
    pw: :obj:`int`
        Plot width including legends, axes etc
    ph: :obj:`int`
        Plot height including legends, axes etc
    x_axis_type: :obj:`str`
        Type of x-axis can be linear, log, or datetime
    y_axis_type: :obj:`str`
        Can be linear, log or datetime
    x_name: :obj:`str`
        Name of the column used for the x-axis. Mostly used to form tooltips
    y_name: :obj:`str`
        Name of the column used for the y-axis. Also used for tooltips
    add_grid: :obj:`bool`
        Whether or not to add grid
    add_title: :obj:`bool`
        Whether or not to add title to plot
    add_xaxis: :obj:`bool`
        Whether or not to add x-axis and tick marks
    add_yaxis: :obj:`bool`
        Add y-axis or not
    fix_plotsize: :obj:`bool`
        Enforce certain dimensions on plot. This is useful for ensuring a plot is not obscure by axes and other things. If activated, plot's dimensions will not be responsive. It utilises fw and fh.

    Returns
    -------
    p : :obj:`Plot`
        A bokeh Plot object

    """

    add_grid = kwargs.pop("add_grid", False)
    add_title = kwargs.pop("add_title", True)
    add_xaxis = kwargs.pop("add_xaxis", False)
    add_yaxis = kwargs.pop("add_yaxis", False)
    fix_plotsize = kwargs.pop("fix_plotsize", True)
    # addition plot specs
    pl_specs = kwargs.pop("pl_specs", {})
    # additional axis specs
    ax_specs = kwargs.pop("ax_specs", {})
    # ticker specs
    ti_specs = kwargs.pop("ti_specs", {})

    plot_specs = dict(background="white", border_fill_alpha=0.1,
                      border_fill_color="white", min_border=3,
                      name="plot", outline_line_dash="solid",
                      outline_line_width=2, outline_line_color="#017afe",
                      outline_line_alpha=0.4, output_backend="canvas",
                      sizing_mode="stretch_width", title_location="above",
                      toolbar_location="right")
    plot_specs.update(pl_specs)

    axis_specs = dict(minor_tick_line_alpha=0, axis_label_text_align="center",
                      axis_label_text_font="monospace",
                      axis_label_text_font_size="10px",
                      axis_label_text_font_style="normal",
                      major_label_orientation="horizontal")
    axis_specs.update(ax_specs)

    tick_specs = dict(desired_num_ticks=5)
    tick_specs.update(ti_specs)

    # Define frame width and height
    # This is the actual size of the plot without the titles et al
    if fix_plotsize and not(fh or fw):
        fw = int(0.98 * pw)
        fh = int(0.93 * ph)

    # percentage increase in axis ranges
    inc = 1.00

    # define the axes ranges
    x_range = DataRange1d(name="p_x_range", only_visible=True)

    y_range = DataRange1d(name="p_y_range", only_visible=True)

    if x_min != None and x_max != None and x_name.lower() in ["channel", "frequency"]:
        x_range = Range1d(name="p_x_range", start=x_min, end=x_max)
        y_range.only_visible = False

    # define items to add on the plot
    p_htool = HoverTool(tooltips=[(x_name, "$x"),
                                  (y_name, "$y")],
                        name="p_htool", point_policy="snap_to_data")

    p_toolbar = Toolbar(name="p_toolbar",
                        tools=[p_htool, BoxSelectTool(), BoxZoomTool(),
                               # EditTool(), # BoxEditTool(), # RangeTool(),
                               LassoSelectTool(), PanTool(), ResetTool(),
                               SaveTool(), UndoTool(), WheelZoomTool()])
    p_ticker = BasicTicker(name="p_ticker", **tick_specs)

    # select the axis scales for x and y
    if x_axis_type == "linear":
        x_scale = LinearScale(name="p_x_scale")
        # define the axes and tickers
        p_x_axis = LinearAxis(axis_label=xlab, name="p_x_axis",
                              ticker=p_ticker, **axis_specs)
    elif x_axis_type == "datetime":
        x_scale = LinearScale(name="p_x_scale")
        # define the axes and tickers
        p_x_axis = DatetimeAxis(axis_label=xlab, name="p_x_axis",
                                ticker=p_ticker, **axis_specs)
    elif x_axis_type == "log":
        x_scale = LogScale(name="p_x_scale")
        p_x_axis = LogAxis(axis_label=xlab, name="p_x_axis",
                           ticker=p_ticker, **axis_specs)

    if y_axis_type == "linear":
        y_scale = LinearScale(name="p_y_scale")
        # define the axes and tickers
        p_y_axis = LinearAxis(axis_label=ylab, name="p_y_axis",
                              ticker=p_ticker, **axis_specs)
    elif x_axis_type == "datetime":
        y_scale = LinearScale(name="p_y_scale")
        # define the axes and tickers
        p_y_axis = DatetimeAxis(axis_label=xlab, name="p_y_axis",
                                ticker=p_ticker, **axis_specs)
    elif y_axis_type == "log":
        y_scale = LogScale(name="p_y_scale")
        # define the axes and tickers
        p_y_axis = LogAxis(axis_label=ylab, name="p_y_axis",
                           ticker=p_ticker, **axis_specs)

    # Create the plot object
    p = Plot(plot_width=pw, plot_height=ph, frame_height=fh, frame_width=fw,
             toolbar=p_toolbar, x_range=x_range, x_scale=x_scale,
             y_range=y_range, y_scale=y_scale, **plot_specs)

    if add_title:
        p_title = Title(align="center", name="p_title", text=title,
                        text_font_size="24px",
                        text_font="monospace", text_font_style="bold",)
        p.add_layout(p_title, "above")

    if add_xaxis:
        p.add_layout(p_x_axis, "below")

    if add_yaxis:
        p.add_layout(p_y_axis, "left")

    if add_grid:
        p_x_grid = Grid(dimension=0, ticker=p_ticker)
        p_y_grid = Grid(dimension=1, ticker=p_ticker)
        p.add_layout(p_x_grid)
        p.add_layout(p_y_grid)

    return p


def add_axis(fig, axis_range, ax_label, ax_name=None):
    """Add an extra axis to the current figure

    Parameters
    ----------
    fig: :obj:`bokeh.plotting.figure`
        The figure onto which to add extra axis

    axis_range: :obj:`list`, :obj:`tuple`
        A range of sorted values or a tuple with 2 values containing or the
        order (min, max). This can be any ordered iteraable that can be
        indexed.

    Returns
    -------
    fig : :obj:`bokeh.plotting.figure`
        Bokeh figure with an extra axis added
    """
    if fig.extra_x_ranges is None:
        fig.extra_x_ranges = {}

    if ax_name is None:
        ax_name = "p_extra_xaxis"

    fig.extra_x_ranges[ax_name] = Range1d(start=axis_range[0],
                                          end=axis_range[-1])

    linaxis = LinearAxis(x_range_name=ax_name, axis_label=ax_label,
                         axis_label_text_font="monospace",
                         axis_label_text_font_style="normal",
                         axis_label_text_font_size="8pt",
                         major_tick_out=2,
                         major_tick_in=2,
                         major_label_text_font_size="8pt",
                         minor_tick_line_width=1,
                         major_label_orientation='horizontal',
                         name=ax_name,
                         ticker=BasicTicker(desired_num_ticks=8))
    fig.add_layout(linaxis, 'above')
    return fig
